# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel, ConfigDict, Field, ImportString


class AuthSettings(BaseModel):
    """Authorization-related settings.

    Here you can set auth provider class along with its options.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__AUTH__PROVIDER=data_rentgen.server.providers.auth.dummy_provider.DummyAuthProvider

        # pass access_key.secret_key = "secret" to DummyAuthProviderSettings
        DATA_RENTGEN__AUTH__ACCESS_KEY__SECRET_KEY=secret
    """

    provider: ImportString = Field(  # type: ignore[assignment]
        default="data_rentgen.server.providers.auth.dummy_provider.DummyAuthProvider",
        description="Full name of auth provider class",
        validate_default=True,
    )

    model_config = ConfigDict(extra="allow")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create user

Revision ID: f9a114951102
Revises: c324e48e0f71
Create Date: 2024-06-27 18:45:00.259710

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f9a114951102"
down_revision = "c324e48e0f71"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "user",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__user")),
    )
    op.create_index(op.f("ix__user__name"), "user", ["name"], unique=True)


def downgrade() -> None:
    op.drop_index(op.f("ix__user__name"), table_name="user")
    op.drop_table("user")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field

from data_rentgen.dto.location import LocationDTO


@dataclass
class DatasetDTO:
    location: LocationDTO
    name: str
    format: str | None = None
    id: int | None = field(default=None, compare=False)

    @property
    def unique_key(self) -> tuple:
        return (self.location.unique_key, self.name)

    def merge(self, new: DatasetDTO) -> DatasetDTO:
        if self.location is new.location and new.id is None and new.format == self.format:
            # datasets aren't changed that much, reuse them if possible
            return self

        return DatasetDTO(
            location=self.location.merge(new.location),
            name=self.name,
            format=new.format or self.format,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.job_facets import OpenLineageJobFacets


class OpenLineageJob(OpenLineageBase):
    """Job model.
    See [Job](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    namespace: str
    name: str
    facets: OpenLineageJobFacets = Field(default_factory=OpenLineageJobFacets)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from enum import IntFlag
from uuid import UUID


class DatasetColumnRelationTypeDTO(IntFlag):
    # See https://github.com/OpenLineage/OpenLineage/blob/1.27.0/integration/spark/shared/src/main/java/io/openlineage/spark/agent/lifecycle/plan/column/TransformationInfo.java#L30-L40
    # Using IntFlag to avoid messing up with ARRAY type, bitwise OR is enough
    UNKNOWN = 1

    # Direct
    IDENTITY = 2
    TRANSFORMATION = 4
    TRANSFORMATION_MASKING = 8
    AGGREGATION = 16
    AGGREGATION_MASKING = 32

    # Indirect
    FILTER = 64
    JOIN = 128
    GROUP_BY = 256
    SORT = 512
    WINDOW = 1024
    CONDITIONAL = 2048


@dataclass
class DatasetColumnRelationDTO:
    type: DatasetColumnRelationTypeDTO
    source_column: str
    target_column: str | None = None
    # id is generated using other ids combination
    fingerprint: UUID | None = None
    # description is always "", see io.openlineage.spark.agent.lifecycle.plan.column.TransformationInfo

    @property
    def unique_key(self) -> tuple:
        return (
            self.source_column,
            self.target_column or "",
        )

    def merge(self, new: DatasetColumnRelationDTO) -> DatasetColumnRelationDTO:
        if new.fingerprint is None and new.type.value & self.type.value:
            return self

        return DatasetColumnRelationDTO(
            source_column=self.source_column,
            target_column=self.target_column,
            type=self.type | new.type,
            fingerprint=new.fingerprint or self.fingerprint,
        )


def merge_dataset_column_relations(
    relations: list[DatasetColumnRelationDTO],
) -> list[DatasetColumnRelationDTO]:
    result: dict[tuple, DatasetColumnRelationDTO] = {}
    for relation in relations:
        if relation.unique_key in result:
            existing_relation = result[relation.unique_key]
            result[relation.unique_key] = existing_relation.merge(relation)
        else:
            result[relation.unique_key] = relation

    return sorted(result.values(), key=lambda item: item.unique_key)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.extractors.batch import BatchExtractionResult, extract_batch
from data_rentgen.consumer.extractors.column_lineage import extract_column_lineage
from data_rentgen.consumer.extractors.dataset import (
    connect_dataset_with_symlinks,
    extract_dataset,
    extract_dataset_and_symlinks,
)
from data_rentgen.consumer.extractors.input import extract_input
from data_rentgen.consumer.extractors.job import extract_job, extract_parent_job
from data_rentgen.consumer.extractors.operation import extract_operation
from data_rentgen.consumer.extractors.output import extract_output
from data_rentgen.consumer.extractors.run import extract_parent_run, extract_run
from data_rentgen.consumer.extractors.schema import extract_schema

__all__ = [
    "BatchExtractionResult",
    "connect_dataset_with_symlinks",
    "extract_batch",
    "extract_column_lineage",
    "extract_dataset",
    "extract_dataset_and_symlinks",
    "extract_input",
    "extract_job",
    "extract_operation",
    "extract_output",
    "extract_parent_job",
    "extract_parent_run",
    "extract_run",
    "extract_schema",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.base import OpenLineageBase


class OpenLineageDatasetFacet(OpenLineageBase):
    """Base class for all dataset facets.
    See [DatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """


class OpenLineageInputDatasetFacet(OpenLineageDatasetFacet):
    """Base class for input dataset facets.
    See [DatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """


class OpenLineageOutputDatasetFacet(OpenLineageDatasetFacet):
    """Base class for output dataset facets.
    See [DatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create location

Revision ID: e730f34b9893
Revises:
Create Date: 2024-06-27 18:44:49.259710

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e730f34b9893"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "location",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("type", sa.String(length=32), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("external_id", sa.String(), nullable=True),
        sa.Column(
            "search_vector",
            postgresql.TSVECTOR(),
            sa.Computed(
                "to_tsvector('simple'::regconfig, name || ' ' || (translate(name, '/.', ' ')) || ' ' || type)",
                persisted=True,
            ),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__location")),
    )
    op.create_index(op.f("ix__location__type_name"), "location", ["type", "name"], unique=True)
    op.create_index(
        op.f("ix__location__search_vector"),
        "location",
        ["search_vector"],
        unique=False,
        postgresql_using="gin",
    )


def downgrade() -> None:
    op.drop_index(op.f("ix__location__type_name"), table_name="location")
    op.drop_index(op.f("ix__location__search_vector"), table_name="location", postgresql_using="gin")
    op.drop_table("location")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from dataclasses import dataclass
from typing import Annotated

from fastapi import Depends

from data_rentgen.db.models.dataset import Dataset
from data_rentgen.dto.pagination import PaginationDTO
from data_rentgen.services.uow import UnitOfWork


@dataclass
class DatasetServiceResult:
    id: int
    data: Dataset


class DatasetServicePaginatedResult(PaginationDTO[DatasetServiceResult]):
    pass


class DatasetService:
    def __init__(self, uow: Annotated[UnitOfWork, Depends()]):
        self._uow = uow

    async def paginate(
        self,
        page: int,
        page_size: int,
        dataset_ids: Collection[int],
        search_query: str | None,
    ) -> DatasetServicePaginatedResult:
        pagination = await self._uow.dataset.paginate(
            page=page,
            page_size=page_size,
            dataset_ids=dataset_ids,
            search_query=search_query,
        )

        return DatasetServicePaginatedResult(
            page=pagination.page,
            page_size=pagination.page_size,
            total_count=pagination.total_count,
            items=[DatasetServiceResult(id=dataset.id, data=dataset) for dataset in pagination.items],
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from data_rentgen.server.services.dataset import DatasetService
from data_rentgen.server.services.get_user import get_user
from data_rentgen.server.services.job import JobService
from data_rentgen.server.services.lineage import LineageService
from data_rentgen.server.services.location import LocationService
from data_rentgen.server.services.operation import OperationService
from data_rentgen.server.services.run import RunService

__all__ = [
    "DatasetService",
    "JobService",
    "LineageService",
    "LocationService",
    "OperationService",
    "RunService",
    "get_user",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Annotated

from packaging.version import Version as RawVersion
from pydantic import PlainSerializer, PlainValidator, WithJsonSchema
from typing_extensions import Doc

from data_rentgen.consumer.openlineage.run_facets.base import OpenLineageRunFacet


def validate_version(value) -> RawVersion:
    if not isinstance(value, RawVersion):
        return RawVersion(value)
    return value


Version = Annotated[
    RawVersion,
    PlainValidator(validate_version),
    PlainSerializer(str, return_type=str),
    WithJsonSchema({"type": "string"}),
    Doc("Version"),
]


class OpenLineageProcessingEngineName(str, Enum):
    AIRFLOW = "Airflow"
    SPARK = "spark"

    def __str__(self) -> str:
        return self.value


class OpenLineageProcessingEngineRunFacet(OpenLineageRunFacet):
    """Run facet describing processing engine.
    See [ProcessingEngineRunFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/ProcessingEngineRunFacet.json).
    """

    name: OpenLineageProcessingEngineName
    version: Version
    openlineageAdapterVersion: Version



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Callable, Coroutine
from typing import Annotated, Any

from fastapi import Depends, Request
from fastapi.security import OAuth2PasswordBearer

from data_rentgen.db.models import User
from data_rentgen.dependencies import Stub
from data_rentgen.server.providers.auth import AuthProvider

oauth_schema = OAuth2PasswordBearer(tokenUrl="v1/auth/token", auto_error=False)


def get_user() -> Callable[[Request, AuthProvider, str], Coroutine[Any, Any, User]]:
    async def wrapper(
        request: Request,
        auth_provider: Annotated[AuthProvider, Depends(Stub(AuthProvider))],
        access_token: Annotated[str | None, Depends(oauth_schema)],
    ) -> User:
        # keycloak provider patches session and store access_token in cookie,
        # dummy auth stores access_token in "Authorization" header
        access_token = request.session.get("access_token") or access_token
        return await auth_provider.get_current_user(  # type: ignore[return-value]
            access_token=access_token,
            request=request,
        )

    return wrapper



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from data_rentgen.server.errors.schemas.invalid_request import InvalidRequestSchema
from data_rentgen.server.errors.schemas.logout import LogoutErrorSchema
from data_rentgen.server.errors.schemas.not_authorized import (
    NotAuthorizedRedirectSchema,
    NotAuthorizedSchema,
)
from data_rentgen.server.errors.schemas.not_found import NotFoundSchema
from data_rentgen.server.errors.schemas.not_implemented import NotImplementedErrorSchema

__all__ = [
    "InvalidRequestSchema",
    "LogoutErrorSchema",
    "NotAuthorizedRedirectSchema",
    "NotAuthorizedSchema",
    "NotFoundSchema",
    "NotImplementedErrorSchema",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.dataset import OpenLineageDataset
from data_rentgen.dto.schema import SchemaDTO


def extract_schema(dataset: OpenLineageDataset) -> SchemaDTO | None:
    if not dataset.facets.datasetSchema:
        return None

    fields = dataset.facets.datasetSchema.fields
    return SchemaDTO(
        fields=[field.model_dump(exclude_unset=True, exclude_none=True) for field in fields],
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum, IntEnum
from uuid import UUID

from data_rentgen.dto.job import JobDTO
from data_rentgen.dto.user import UserDTO


class RunStatusDTO(IntEnum):
    UNKNOWN = -1
    STARTED = 0
    SUCCEEDED = 1
    FAILED = 2
    KILLED = 3


class RunStartReasonDTO(str, Enum):
    MANUAL = "MANUAL"
    AUTOMATIC = "AUTOMATIC"

    def __str__(self) -> str:
        return str(self.value)


@dataclass
class RunDTO:
    id: UUID
    job: JobDTO
    parent_run: RunDTO | None = None
    status: RunStatusDTO = RunStatusDTO.UNKNOWN
    started_at: datetime | None = None
    start_reason: RunStartReasonDTO | None = None
    user: UserDTO | None = None
    ended_at: datetime | None = None
    external_id: str | None = None
    attempt: str | None = None
    running_log_url: str | None = None
    persistent_log_url: str | None = None

    @property
    def unique_key(self) -> tuple:
        return (self.id,)

    def merge(self, new: RunDTO) -> RunDTO:
        parent_run: RunDTO | None
        if new.parent_run and self.parent_run:
            parent_run = self.parent_run.merge(new.parent_run)
        else:
            parent_run = new.parent_run or self.parent_run

        user: UserDTO | None
        if new.user and self.user:  # noqa: SIM108
            user = self.user.merge(new.user)
        else:
            user = new.user or self.user

        return RunDTO(
            id=self.id,
            job=self.job.merge(new.job),
            parent_run=parent_run,
            status=max(new.status, self.status),
            started_at=new.started_at or self.started_at,
            start_reason=new.start_reason or self.start_reason,
            user=user,
            ended_at=new.ended_at or self.ended_at,
            external_id=new.external_id or self.external_id,
            attempt=new.attempt or self.attempt,
            running_log_url=new.running_log_url or self.running_log_url,
            persistent_log_url=new.persistent_log_url or self.persistent_log_url,
        )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import asyncio
import logging
import sys
from argparse import ArgumentParser
from datetime import UTC, date, datetime
from enum import Enum

from dateutil.parser import isoparse
from dateutil.relativedelta import relativedelta
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from data_rentgen.db.factory import create_session_factory
from data_rentgen.db.models import Input, Operation, Output, Run
from data_rentgen.db.models.column_lineage import ColumnLineage
from data_rentgen.db.settings import DatabaseSettings
from data_rentgen.logging.settings import LoggingSettings
from data_rentgen.logging.setup_logging import setup_logging

logger = logging.getLogger(__name__)

PARTITIONED_TABLES = [
    Run.__tablename__,
    Operation.__tablename__,
    Input.__tablename__,
    Output.__tablename__,
    ColumnLineage.__tablename__,
]


class Granularity(str, Enum):
    DAY = "day"
    MONTH = "month"
    YEAR = "year"

    def __str__(self) -> str:
        return self.value

    def round(self, input_: date) -> date:
        if self == Granularity.DAY:
            return input_
        if self == Granularity.MONTH:
            return input_.replace(day=1)
        return input_.replace(day=1, month=1)

    def to_range(self) -> relativedelta:
        if self == Granularity.DAY:
            return relativedelta(days=1)
        if self == Granularity.MONTH:
            return relativedelta(months=1)
        return relativedelta(years=1)

    def to_format(self) -> str:
        if self == Granularity.DAY:
            return "y%Y_m%m_%d"
        if self == Granularity.MONTH:
            return "y%Y_m%m"
        return "y%Y"


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        usage="python3 -m data_rentgen.db.scripts.create_partitions --start 2024-01 --granularity month",
        description="Create partitions for tables, for given start date and granularity.",
    )
    parser.add_argument(
        "--start",
        type=isoparse,
        default=datetime.now(tz=UTC).replace(day=1, hour=0, minute=0, second=0, microsecond=0),
        nargs="?",
        help="Start date for partitions, default is the first day of current month.",
    )
    parser.add_argument(
        "--end",
        type=isoparse,
        default=datetime.now(tz=UTC).replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        + relativedelta(months=2),
        nargs="?",
        help="End date for partitions, default is the last day of next month.",
    )
    parser.add_argument(
        "--granularity",
        type=lambda x: Granularity(x).value,
        choices=[item.value for item in Granularity],
        default=Granularity.MONTH.value,
        nargs="?",
        help="Granularity of partitions, default is month.",
    )
    return parser


def generate_partition_range(start: date, end: date, granularity: Granularity) -> list[tuple[date, date]]:
    parts = []
    current = start
    next_part = current + granularity.to_range()
    while current < end:
        parts.append((current, next_part))
        current += granularity.to_range()
        next_part = current + granularity.to_range()
    if not parts:
        msg = "Data range is too small"
        raise ValueError(msg)

    return parts


async def create_partition(start: date, end: date, granularity: Granularity, session: AsyncSession):
    partition_name = start.strftime(granularity.to_format())

    start_str = start.isoformat()
    end_str = end.isoformat()

    for table in PARTITIONED_TABLES:
        statement = f"CREATE TABLE IF NOT EXISTS {table}_{partition_name} PARTITION OF {table} FOR VALUES FROM ('{start_str}') TO ('{end_str}')"  # noqa: E501
        logger.debug("Executing statement: %s", statement)
        await session.execute(text(statement))


async def main(args: list[str]) -> None:
    setup_logging(LoggingSettings())

    parser = get_parser()
    params = parser.parse_args(args)
    granularity = Granularity(params.granularity)
    if params.end is None:
        params.end = params.start + granularity.to_range()

    start, end = granularity.round(params.start.date()), granularity.round(params.end.date())
    if start > end:
        msg = "Start date must be less than end date."
        raise ValueError(msg)

    logger.info("Creating partitions from %s to %s with %s granularity", start, end, granularity.value)

    db_settings = DatabaseSettings()  # type: ignore[call-arg]
    session_factory = create_session_factory(db_settings)
    async with session_factory() as session:
        for from_, to in generate_partition_range(start, end, granularity):
            await create_partition(from_, to, granularity, session)
        await session.commit()


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TypeVar

from data_rentgen.consumer.extractors.column_lineage import extract_column_lineage
from data_rentgen.consumer.extractors.input import extract_input
from data_rentgen.consumer.extractors.operation import extract_operation
from data_rentgen.consumer.extractors.output import extract_output
from data_rentgen.consumer.extractors.run import extract_run
from data_rentgen.consumer.openlineage.job_facets.job_type import OpenLineageJobType
from data_rentgen.consumer.openlineage.run_event import OpenLineageRunEvent
from data_rentgen.dto import (
    ColumnLineageDTO,
    DatasetDTO,
    DatasetSymlinkDTO,
    InputDTO,
    JobDTO,
    LocationDTO,
    OperationDTO,
    OutputDTO,
    RunDTO,
    SchemaDTO,
    UserDTO,
)

T = TypeVar(
    "T",
    LocationDTO,
    DatasetDTO,
    ColumnLineageDTO,
    DatasetSymlinkDTO,
    JobDTO,
    RunDTO,
    OperationDTO,
    InputDTO,
    OutputDTO,
    SchemaDTO,
    UserDTO,
)


class BatchExtractionResult:
    """Track results of batch extraction.

    Calling any ``add_*`` method will add DTO item to the result, including nested DTOs,
    like ``OperationDTO`` -> ``RunDTO`` -> ``JobDTO`` -> ``LocationDTO``, and so on.

    Each DTO type is tracked separately. DTOs with same ``unique_key`` are merged into one final DTO,
    by calling ``existing.merge(new)``. The resulting final DTO contains all non-null attributes of original DTOs.
    Last DTO in the chain has a higher priority than previuos ones.
    For example ``RunDTO(status=STARTED, started_at=...).merge(RunDTO(status=SUCCEEDED, ended_at=...))``
    produces final ``RunDTO(status=SUCCEEDED, started_at=..., ended_at=...)``.

    Calling get methods, like ``jobs()``, will return the list of tracked DTOs with resolved
    cross-links. For example, iterating over ``jobs()`` with return the same objects
    as in ``[run.job for run in runs()]``.
    This makes modification of nested DTOs easier, all changes will be reflected in the parent DTOs as well.
    """

    def __init__(self):
        self._locations: dict[tuple, LocationDTO] = {}
        self._datasets: dict[tuple, DatasetDTO] = {}
        self._dataset_symlinks: dict[tuple, DatasetSymlinkDTO] = {}
        self._jobs: dict[tuple, JobDTO] = {}
        self._runs: dict[tuple, RunDTO] = {}
        self._operations: dict[tuple, OperationDTO] = {}
        self._inputs: dict[tuple, InputDTO] = {}
        self._outputs: dict[tuple, OutputDTO] = {}
        self._column_lineage: dict[tuple, ColumnLineageDTO] = {}
        self._schemas: dict[tuple, SchemaDTO] = {}
        self._users: dict[tuple, UserDTO] = {}

    def __repr__(self):
        return (
            "ExtractionResult("
            f"locations={len(self._locations)}, "
            f"datasets={len(self._datasets)}, "
            f"dataset_symlinks={len(self._dataset_symlinks)}, "
            f"jobs={len(self._jobs)}, "
            f"runs={len(self._runs)}, "
            f"operations={len(self._operations)}, "
            f"inputs={len(self._inputs)}, "
            f"outputs={len(self._outputs)}, "
            f"column_lineage={len(self._column_lineage)}, "
            f"schemas={len(self._schemas)}, "
            f"users={len(self._users)}"
            ")"
        )

    @staticmethod
    def _add(context: dict[tuple, T], new_item: T) -> T:
        key = new_item.unique_key
        if key in context:
            old_item = context[key]
            if old_item is new_item:
                return old_item

            merged_item = old_item.merge(new_item)
            context[key] = merged_item
            return merged_item

        context[key] = new_item
        return new_item

    def add_location(self, location: LocationDTO):
        return self._add(self._locations, location)

    def add_dataset(self, dataset: DatasetDTO):
        dataset.location = self.add_location(dataset.location)
        return self._add(self._datasets, dataset)

    def add_dataset_symlink(self, dataset_symlink: DatasetSymlinkDTO):
        dataset_symlink.from_dataset = self.add_dataset(dataset_symlink.from_dataset)
        dataset_symlink.to_dataset = self.add_dataset(dataset_symlink.to_dataset)
        return self._add(self._dataset_symlinks, dataset_symlink)

    def add_job(self, job: JobDTO):
        job.location = self.add_location(job.location)
        return self._add(self._jobs, job)

    def add_run(self, run: RunDTO):
        run.job = self.add_job(run.job)
        if run.parent_run:
            run.parent_run = self.add_run(run.parent_run)
        if run.user:
            run.user = self.add_user(run.user)
        return self._add(self._runs, run)

    def add_operation(self, operation: OperationDTO):
        operation.run = self.add_run(operation.run)
        return self._add(self._operations, operation)

    def add_input(self, input_: InputDTO):
        input_.operation = self.add_operation(input_.operation)
        input_.dataset = self.add_dataset(input_.dataset)
        if input_.schema:
            input_.schema = self.add_schema(input_.schema)
        return self._add(self._inputs, input_)

    def add_output(self, output: OutputDTO):
        output.operation = self.add_operation(output.operation)
        output.dataset = self.add_dataset(output.dataset)
        if output.schema:
            output.schema = self.add_schema(output.schema)
        return self._add(self._outputs, output)

    def add_column_lineage(self, lineage: ColumnLineageDTO):
        lineage.source_dataset = self.add_dataset(lineage.source_dataset)
        lineage.target_dataset = self.add_dataset(lineage.target_dataset)
        lineage.operation = self.add_operation(lineage.operation)
        return self._add(self._column_lineage, lineage)

    def add_schema(self, schema: SchemaDTO):
        return self._add(self._schemas, schema)

    def add_user(self, user: UserDTO):
        return self._add(self._users, user)

    def get_location(self, location_key: tuple) -> LocationDTO:
        return self._locations[location_key]

    def get_schema(self, schema_key: tuple) -> SchemaDTO:
        return self._schemas[schema_key]

    def get_user(self, user_key: tuple) -> UserDTO:
        return self._users[user_key]

    def get_dataset(self, dataset_key: tuple) -> DatasetDTO:
        dataset = self._datasets[dataset_key]
        dataset.location = self.get_location(dataset.location.unique_key)
        return dataset

    def get_dataset_symlink(self, dataset_symlink_key: tuple) -> DatasetSymlinkDTO:
        dataset_symlink = self._dataset_symlinks[dataset_symlink_key]
        dataset_symlink.from_dataset = self.get_dataset(dataset_symlink.from_dataset.unique_key)
        dataset_symlink.to_dataset = self.get_dataset(dataset_symlink.to_dataset.unique_key)
        return dataset_symlink

    def get_job(self, job_key: tuple) -> JobDTO:
        job = self._jobs[job_key]
        job.location = self.get_location(job.location.unique_key)
        return job

    def get_run(self, run_key: tuple) -> RunDTO:
        run = self._runs[run_key]
        run.job = self.get_job(run.job.unique_key)
        if run.parent_run:
            run.parent_run = self.get_run(run.parent_run.unique_key)
        if run.user:
            run.user = self.get_user(run.user.unique_key)
        return run

    def get_operation(self, operation_key: tuple) -> OperationDTO:
        operation = self._operations[operation_key]
        operation.run = self.get_run(operation.run.unique_key)
        return operation

    def get_input(self, input_key: tuple) -> InputDTO:
        input_ = self._inputs[input_key]
        input_.operation = self.get_operation(input_.operation.unique_key)
        input_.dataset = self.get_dataset(input_.dataset.unique_key)
        if input_.schema:
            input_.schema = self.get_schema(input_.schema.unique_key)
        return input_

    def get_output(self, output_key: tuple) -> OutputDTO:
        output = self._outputs[output_key]
        output.operation = self.get_operation(output.operation.unique_key)
        output.dataset = self.get_dataset(output.dataset.unique_key)
        if output.schema:
            output.schema = self.get_schema(output.schema.unique_key)
        return output

    def get_column_lineage(self, output_key: tuple) -> ColumnLineageDTO:
        lineage = self._column_lineage[output_key]
        lineage.operation = self.get_operation(lineage.operation.unique_key)
        lineage.source_dataset = self.get_dataset(lineage.source_dataset.unique_key)
        lineage.target_dataset = self.get_dataset(lineage.target_dataset.unique_key)
        return lineage

    def locations(self) -> list[LocationDTO]:
        return list(map(self.get_location, self._locations))

    def datasets(self) -> list[DatasetDTO]:
        return list(map(self.get_dataset, self._datasets))

    def dataset_symlinks(self) -> list[DatasetSymlinkDTO]:
        return list(map(self.get_dataset_symlink, self._dataset_symlinks))

    def jobs(self) -> list[JobDTO]:
        return list(map(self.get_job, self._jobs))

    def runs(self) -> list[RunDTO]:
        return list(map(self.get_run, self._runs))

    def operations(self) -> list[OperationDTO]:
        return list(map(self.get_operation, self._operations))

    def inputs(self) -> list[InputDTO]:
        return list(map(self.get_input, self._inputs))

    def outputs(self) -> list[OutputDTO]:
        return list(map(self.get_output, self._outputs))

    def column_lineage(self) -> list[ColumnLineageDTO]:
        return list(map(self.get_column_lineage, self._column_lineage))

    def schemas(self) -> list[SchemaDTO]:
        return list(map(self.get_schema, self._schemas))

    def users(self) -> list[UserDTO]:
        return list(map(self.get_user, self._users))


def extract_batch(events: list[OpenLineageRunEvent]) -> BatchExtractionResult:
    result = BatchExtractionResult()
    dataset_cache: dict[tuple[str, str], DatasetDTO] = {}

    for event in events:
        if event.job.facets.jobType and event.job.facets.jobType.jobType == OpenLineageJobType.JOB:
            operation = extract_operation(event)
            result.add_operation(operation)

            for input_dataset in event.inputs:
                input_dto, symlink_dtos = extract_input(operation, input_dataset)

                result.add_input(input_dto)
                dataset_dto_cache_key = (input_dataset.namespace, input_dataset.name)
                dataset_cache[dataset_dto_cache_key] = result.get_dataset(input_dto.dataset.unique_key)

                for symlink_dto in symlink_dtos:
                    result.add_dataset_symlink(symlink_dto)

            for output_dataset in event.outputs:
                output_dto, symlink_dtos = extract_output(operation, output_dataset)

                result.add_output(output_dto)
                dataset_dto_cache_key = (output_dataset.namespace, output_dataset.name)
                dataset_cache[dataset_dto_cache_key] = result.get_dataset(output_dto.dataset.unique_key)

                for symlink_dto in symlink_dtos:
                    result.add_dataset_symlink(symlink_dto)

            for dataset in event.inputs + event.outputs:
                column_lineage = extract_column_lineage(operation, dataset, dataset_cache)
                for item in column_lineage:
                    result.add_column_lineage(item)

        else:
            run = extract_run(event)
            result.add_run(run)

    return result



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import defaultdict

from data_rentgen.consumer.extractors.dataset import extract_dataset_ref
from data_rentgen.consumer.openlineage.dataset import OpenLineageDataset
from data_rentgen.consumer.openlineage.dataset_facets.column_lineage import (
    OpenLineageColumnLineageDatasetFacetFieldRef,
    OpenLineageColumnLineageDatasetFacetFieldTransformation,
)
from data_rentgen.dto import (
    ColumnLineageDTO,
    DatasetColumnRelationDTO,
    DatasetColumnRelationTypeDTO,
)
from data_rentgen.dto.dataset import DatasetDTO
from data_rentgen.dto.operation import OperationDTO

logger = logging.getLogger(__name__)

TRANSFORMATION_TYPE_DIRECT = "DIRECT"

TRANSFORMATION_SUBTYPE_MAP_MASKING = {
    "TRANSFORMATION": DatasetColumnRelationTypeDTO.TRANSFORMATION_MASKING,
    "AGGREGATION": DatasetColumnRelationTypeDTO.AGGREGATION_MASKING,
}

TRANSFORMATION_SUBTYPE_MAP = {
    "IDENTITY": DatasetColumnRelationTypeDTO.IDENTITY,
    "TRANSFORMATION": DatasetColumnRelationTypeDTO.TRANSFORMATION,
    "AGGREGATION": DatasetColumnRelationTypeDTO.AGGREGATION,
    "FILTER": DatasetColumnRelationTypeDTO.FILTER,
    "JOIN": DatasetColumnRelationTypeDTO.JOIN,
    "GROUP_BY": DatasetColumnRelationTypeDTO.GROUP_BY,
    "SORT": DatasetColumnRelationTypeDTO.SORT,
    "WINDOW": DatasetColumnRelationTypeDTO.WINDOW,
    "CONDITIONAL": DatasetColumnRelationTypeDTO.CONDITIONAL,
}


def extract_dataset_column_relation_type(
    transformation: OpenLineageColumnLineageDatasetFacetFieldTransformation,
) -> DatasetColumnRelationTypeDTO:
    result: DatasetColumnRelationTypeDTO | None = None
    if transformation.subtype:
        if transformation.masking:
            result = TRANSFORMATION_SUBTYPE_MAP_MASKING.get(transformation.subtype)
        else:
            result = TRANSFORMATION_SUBTYPE_MAP.get(transformation.subtype)

    return result or DatasetColumnRelationTypeDTO.UNKNOWN


def resolve_dataset_ref(
    dataset_ref: OpenLineageDataset | OpenLineageColumnLineageDatasetFacetFieldRef,
    dataset_dto_cache: dict[tuple[str, str], DatasetDTO],
):
    # extracting dataset for every column is expensive. cache it as much as we can
    dataset_cache_key = (dataset_ref.namespace, dataset_ref.name)
    if dataset_cache_key not in dataset_dto_cache:
        # https://github.com/OpenLineage/OpenLineage/issues/2938#issuecomment-2320377260
        dataset_dto_cache[dataset_cache_key] = extract_dataset_ref(dataset_ref)
    return dataset_dto_cache[dataset_cache_key]


def extract_column_lineage(
    operation: OperationDTO,
    target_dataset: OpenLineageDataset,
    dataset_cache: dict[tuple[str, str], DatasetDTO] | None = None,
) -> list[ColumnLineageDTO]:
    if not target_dataset.facets.columnLineage:
        return []

    dataset_cache = dataset_cache or {}
    target_dataset_dto = resolve_dataset_ref(target_dataset, dataset_cache)

    # Grouping column lineage by source+target dataset. This is unique combination within operation,
    # so we can use it to generate the same fingerprint for all dataset column relations
    datasets = {target_dataset_dto.unique_key: target_dataset_dto}
    dataset_column_relations: dict[tuple, dict[tuple, DatasetColumnRelationDTO]] = defaultdict(dict)

    # direct lineage (source_column -> target_column)
    for field, raw_column_lineage in target_dataset.facets.columnLineage.fields.items():
        for input_field in raw_column_lineage.inputFields:
            source_dataset_dto = resolve_dataset_ref(input_field, dataset_cache)
            datasets[source_dataset_dto.unique_key] = source_dataset_dto

            dataset_relation_key = (source_dataset_dto.unique_key, target_dataset_dto.unique_key)
            dataset_column_relation = dataset_column_relations[dataset_relation_key]

            for transformation in input_field.transformations:
                # OL integration for Spark before v1.23
                # or with columnLineage.datasetLineageEnabled=false (which is still default)
                # produces INDIRECT lineage for each combination source_column x target_column,
                # which is almost a cartesian product.
                # There are a lot of duplicates here, trying to avoid them by merging items immediately.
                column_relation = DatasetColumnRelationDTO(
                    type=extract_dataset_column_relation_type(transformation),
                    source_column=input_field.field,
                    target_column=field if transformation.type == TRANSFORMATION_TYPE_DIRECT else None,
                )
                column_relation_key = column_relation.unique_key

                existing_column_relation = dataset_column_relation.get(column_relation_key)
                if existing_column_relation:
                    dataset_column_relation[column_relation_key] = existing_column_relation.merge(column_relation)
                else:
                    dataset_column_relation[column_relation_key] = column_relation

    # indirect lineage (source_column -> target_dataset),
    # added to OL since v1.23 and send only when columnLineage.datasetLineageEnabled=true
    for input_field in target_dataset.facets.columnLineage.dataset:
        source_dataset_dto = resolve_dataset_ref(input_field, dataset_cache)
        datasets[source_dataset_dto.unique_key] = source_dataset_dto

        dataset_relation_key = (source_dataset_dto.unique_key, target_dataset_dto.unique_key)
        dataset_column_relation = dataset_column_relations[dataset_relation_key]

        for transformation in input_field.transformations:
            column_relation = DatasetColumnRelationDTO(
                type=extract_dataset_column_relation_type(transformation),
                source_column=input_field.field,
            )
            column_relation_key = column_relation.unique_key

            existing_column_relation = dataset_column_relation.get(column_relation_key)
            if existing_column_relation:
                dataset_column_relation[column_relation_key] = existing_column_relation.merge(column_relation)
            else:
                dataset_column_relation[column_relation_key] = column_relation

    # merge results into DTO objects
    return [
        ColumnLineageDTO(
            operation=operation,
            source_dataset=datasets[source_dataset_dto_key],
            target_dataset=datasets[target_dataset_dto_key],
            dataset_column_relations=list(relations.values()),
        )
        for (source_dataset_dto_key, target_dataset_dto_key), relations in dataset_column_relations.items()
        if dataset_column_relations
    ]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Generic, TypeVar

from pydantic import BaseModel


class BaseErrorSchema(BaseModel):
    code: str
    message: str
    details: Any


ErrorModel = TypeVar("ErrorModel", bound=BaseErrorSchema)


class APIErrorSchema(BaseModel, Generic[ErrorModel]):
    error: ErrorModel



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends

from data_rentgen.db.models import User
from data_rentgen.server.errors import get_error_responses
from data_rentgen.server.errors.schemas import InvalidRequestSchema, NotFoundSchema
from data_rentgen.server.schemas.v1 import (
    LocationDetailedResponseV1,
    LocationPaginateQueryV1,
    PageResponseV1,
    UpdateLocationRequestV1,
)
from data_rentgen.server.services import LocationService, get_user

router = APIRouter(
    prefix="/locations",
    tags=["Locations"],
    responses=get_error_responses(include={InvalidRequestSchema}),
)


@router.get("", summary="Paginated list of Locations")
async def paginate_locations(
    query_args: Annotated[LocationPaginateQueryV1, Depends()],
    location_service: Annotated[LocationService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> PageResponseV1[LocationDetailedResponseV1]:
    pagination = await location_service.paginate(
        page=query_args.page,
        page_size=query_args.page_size,
        location_ids=query_args.location_id,
        location_type=query_args.location_type,
        search_query=query_args.search_query,
    )
    return PageResponseV1[LocationDetailedResponseV1].from_pagination(pagination)


@router.patch(
    "/{location_id}",
    responses=get_error_responses(include={InvalidRequestSchema, NotFoundSchema}),
)
async def update_location(
    location_id: int,
    location_data: UpdateLocationRequestV1,
    location_service: Annotated[LocationService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> LocationDetailedResponseV1:
    location = await location_service.update_external_id(location_id, location_data.external_id)
    return LocationDetailedResponseV1.model_validate(location)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import json
from typing import Any
from uuid import UUID

from data_rentgen.db.utils.uuid import generate_static_uuid


def get_fields_digest(fields: Any) -> UUID:
    """Use this function to generate SHA-1 digest of schema fields"""
    content = json.dumps(fields, sort_keys=True)
    return generate_static_uuid(content)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import textwrap
from typing import Literal

from pydantic import BaseModel, ByteSize, Field, model_validator


class ConsumerSettings(BaseModel):
    """Data.Rentgen consumer-specific settings.

    These options are passed directly to
    `AIOKafkaConsumer <https://aiokafka.readthedocs.io/en/stable/api.html#aiokafka.AIOKafkaConsumer>`_.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__CONSUMER__TOPICS_LIST=["input.runs"]
        DATA_RENTGEN__CONSUMER__GROUP_ID=data-rentgen
        DATA_RENTGEN__CONSUMER__FETCH_MAX_WAIT_MS=5000
        DATA_RENTGEN__CONSUMER__MAX_PARTITION_FETCH_BYTES=5MiB
    """

    topics_list: list[str] = Field(
        default=["input.runs"],
        description="List of Kafka topics to subscribe. Mutually exclusive with :obj:`~topics_pattern`.",
    )
    topics_pattern: str | None = Field(
        default=None,
        description="Regex pattern of topics to subscribe. Mutually exclusive with :obj:`~topics_list`.",
    )

    @model_validator(mode="after")
    def _check_topics(self):
        if not self.topics_list and not self.topics_pattern:
            msg = "input should contain either 'topics_list' or 'topics_pattern' field, both are empty"
            raise ValueError(msg)
        if self.topics_list and self.topics_pattern:
            msg = "input should contain either 'topics_list' or 'topics_pattern' field, both are set"
            raise ValueError(msg)
        return self

    group_id: str | None = Field(
        default="data-rentgen",
        description=textwrap.dedent(
            """
            Name of the consumer group to join for dynamic partition assignment (if enabled),
            and to use for fetching and committing offsets.
            If ``None``, auto-partition assignment (via group coordinator) and offset commits are disabled.
            """,
        ),
    )

    # Defaults are copied from FastStream: https://github.com/airtai/faststream/blob/0.5.33/faststream/kafka/fastapi/fastapi.py#L618
    # But only options, related to consumer
    max_records: int | None = Field(
        default=None,
        description=textwrap.dedent(
            """
            Number of messages to consume as one batch.
            ``None`` means no limit applied.
            """,
        ),
    )
    fetch_max_bytes: ByteSize = Field(
        default=ByteSize(50 * 1024 * 1024),
        description="The maximum amount of data the server should return for a fetch request.",
    )
    fetch_min_bytes: ByteSize = Field(
        default=ByteSize(1),
        description=textwrap.dedent(
            """
            Minimum amount of data the server should
            return for a fetch request, otherwise wait up to
            :obj:`~fetch_max_wait_ms` for more data to accumulate.
            """,
        ),
    )
    fetch_max_wait_ms: int = Field(
        default=500,
        description=textwrap.dedent(
            """
            The maximum amount of time in milliseconds
            the server will block before answering the fetch request if
            there isn't sufficient data to immediately satisfy the
            requirement given by :obj:`~fetch_min_bytes`.
            """,
        ),
    )
    max_partition_fetch_bytes: ByteSize = Field(
        default=ByteSize(1024 * 1024),
        description=textwrap.dedent(
            """
            The maximum amount of data
            per-partition the server will return. The maximum total memory
            used for a request ``= #partitions * max_partition_fetch_bytes``.

            This size must be at least as large as the maximum message size
            the server allows or else it is possible for the producer to
            send messages larger than the consumer can fetch. If that
            happens, the consumer can get stuck trying to fetch a large
            message on a certain partition.
            """,
        ),
    )
    batch_timeout_ms: int = Field(
        default=200,
        description=textwrap.dedent(
            """
            Milliseconds spent waiting if data is not available in the buffer.
            If 0, returns immediately with any records that are available currently in the buffer,
            else returns empty.
            """,
        ),
    )
    auto_offset_reset: Literal["latest", "earliest", "none"] = Field(
        default="latest",
        description=textwrap.dedent(
            """"
            A policy for resetting offsets on ``OffsetOutOfRangeError`` errors:

            * ``earliest`` will move to the oldest available message
            * ``latest`` will move to the most recent
            * ``none`` will raise an exception so you can handle this case
            """,
        ),
    )
    max_poll_interval_ms: int = Field(
        default=5 * 60 * 1000,
        description=textwrap.dedent(
            """
            Maximum allowed time between calls to consume messages in batches.
            If this interval is exceeded the consumer is considered failed and the group will
            rebalance in order to reassign the partitions to another consumer
            group member.
            If API methods block waiting for messages, that time
            does not count against this timeout.
            """,
        ),
    )
    session_timeout_ms: int = Field(
        default=10 * 1000,
        description=textwrap.dedent(
            """
            Client group session and failure detection
            timeout. The consumer sends periodic heartbeats
            (``heartbeat.interval.ms``) to indicate its liveness to the broker.

            If no hearts are received by the broker for a group member within
            the session timeout, the broker will remove the consumer from the
            group and trigger a rebalance.

            The allowed range is configured with the **broker** configuration properties
            ``group.min.session.timeout.ms`` and ``group.max.session.timeout.ms``.
            """,
        ),
    )
    heartbeat_interval_ms: int = Field(
        default=3 * 1000,
        description=textwrap.dedent(
            """
            The expected time in milliseconds
            between heartbeats to the consumer coordinator when using
            Kafka's group management feature. Heartbeats are used to ensure
            that the consumer's session stays active and to facilitate
            rebalancing when new consumers join or leave the group.

            The value must be set lower than :obj:`~session_timeout_ms`, but typically
            should be set no higher than 1/3 of that value. It can be
            adjusted even lower to control the expected time for normal
            rebalances.
            """,
        ),
    )
    consumer_timeout_ms: int = Field(
        default=200,
        description=textwrap.dedent(
            """
            Maximum wait timeout for background fetching routine.
            Mostly defines how fast the system will see rebalance and
            request new data for new partitions.
            """,
        ),
    )
    isolation_level: Literal["read_uncommitted", "read_committed"] = Field(
        default="read_uncommitted",
        description=textwrap.dedent(
            """
            Controls how to read messages written
            transactionally.

            * ``read_committed`` - batch consumer will only return
              transactional messages which have been committed.

            * ``read_uncommitted`` (the default) - batch consumer will
              return all messages, even transactional messages which have been
              aborted.

            Non-transactional messages will be returned unconditionally in
            either mode.

            Messages will always be returned in offset order. Hence, in
            ``read_committed`` mode, batch consumer will only return
            messages up to the last stable offset (LSO), which is the one less
            than the offset of the first open transaction. In particular any
            messages appearing after messages belonging to ongoing transactions
            will be withheld until the relevant transaction has been completed.
            As a result, ``read_committed`` consumers will not be able to read up
            to the high watermark when there are in flight transactions.
            Further, when in ``read_committed`` the seek_to_end method will
            return the LSO.
            """,
        ),
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends, Query

from data_rentgen.db.models import User
from data_rentgen.server.errors import get_error_responses
from data_rentgen.server.errors.schemas import InvalidRequestSchema
from data_rentgen.server.schemas.v1 import (
    LineageResponseV1,
    PageResponseV1,
    RunDetailedResponseV1,
    RunLineageQueryV1,
    RunsQueryV1,
)
from data_rentgen.server.services import LineageService, RunService, get_user
from data_rentgen.server.utils.lineage_response import build_lineage_response

router = APIRouter(prefix="/runs", tags=["Runs"], responses=get_error_responses(include={InvalidRequestSchema}))


@router.get("", summary="Paginated list of Runs")
async def runs(
    query_args: Annotated[RunsQueryV1, Depends()],
    run_service: Annotated[RunService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> PageResponseV1[RunDetailedResponseV1]:
    pagination = await run_service.paginate(
        page=query_args.page,
        page_size=query_args.page_size,
        since=query_args.since,
        until=query_args.until,
        run_ids=query_args.run_id,
        job_id=query_args.job_id,
        parent_run_id=query_args.parent_run_id,
        search_query=query_args.search_query,
    )
    return PageResponseV1[RunDetailedResponseV1].from_pagination(pagination)


@router.get("/lineage", summary="Get Run lineage graph")
async def get_runs_lineage(
    query_args: Annotated[RunLineageQueryV1, Query()],
    lineage_service: Annotated[LineageService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> LineageResponseV1:
    lineage = await lineage_service.get_lineage_by_runs(
        start_node_ids=[query_args.start_node_id],
        direction=query_args.direction,
        granularity=query_args.granularity,
        since=query_args.since,
        until=query_args.until,
        depth=query_args.depth,
        include_column_lineage=query_args.include_column_lineage,
    )
    return await build_lineage_response(lineage)



# SPDX-FileCopyrightText: 2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create custom properties

Revision ID: 61ea5edad711
Revises: 3357c8f914aa
Create Date: 2025-01-09 12:12:04.489277

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "61ea5edad711"
down_revision = "3357c8f914aa"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "custom_properties",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__custom_properties")),
        sa.UniqueConstraint("name", name=op.f("uq__custom_properties__name")),
    )


def downgrade() -> None:
    op.drop_table("custom_properties")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from data_rentgen.db.settings import DatabaseSettings
from data_rentgen.logging.settings import LoggingSettings
from data_rentgen.server.settings.auth import AuthSettings
from data_rentgen.server.settings.server import ServerSettings


class ServerApplicationSettings(BaseSettings):
    """Data.Rentgen REST API settings.

    Application can be configured in 2 ways:

    * By explicitly passing ``settings`` object as an argument to :obj:`application_factory <data_rentgen.server.main.application_factory>`
    * By setting up environment variables matching a specific key.

        All environment variable names are written in uppercase and should be prefixed with ``DATA_RENTGEN__``.
        Nested items are delimited with ``__``.

    More details can be found in `Pydantic documentation <https://docs.pydantic.dev/latest/concepts/pydantic_settings/>`_.

    Examples
    --------

    .. code-block:: bash

        # same as settings.database.url = "postgresql+asyncpg://postgres:postgres@localhost:5432/data_rentgen"
        DATA_RENTGEN__DATABASE__URL=postgresql+asyncpg://postgres:postgres@localhost:5432/data_rentgen

        # same as settings.logging.preset = "json"
        DATA_RENTGEN__LOGGING__PRESET=json

        # same as settings.server.debug = True
        DATA_RENTGEN__SERVER__DEBUG=True
    """  # noqa: E501

    auth: AuthSettings = Field(
        default_factory=AuthSettings,
        description="Auth settings",
    )
    database: DatabaseSettings = Field(description=":ref:`Database settings <configuration-database>`")
    logging: LoggingSettings = Field(
        default_factory=LoggingSettings,
        description=":ref:`Logging settings <configuration-server-logging>`",
    )
    server: ServerSettings = Field(
        default_factory=ServerSettings,
        description=":ref:`Server settings <configuration-server>`",
    )

    model_config = SettingsConfigDict(env_prefix="DATA_RENTGEN__", env_nested_delimiter="__", extra="forbid")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import logging
from pprint import pformat
from time import time
from typing import Annotated, Any

from fastapi import Depends, FastAPI, Request

from data_rentgen.db.models import User
from data_rentgen.dependencies import Stub
from data_rentgen.dto import UserDTO
from data_rentgen.exceptions import EntityNotFoundError
from data_rentgen.exceptions.auth import AuthorizationError
from data_rentgen.server.providers.auth.base_provider import AuthProvider
from data_rentgen.server.settings.auth.dummy import DummyAuthProviderSettings
from data_rentgen.server.utils.jwt import decode_jwt, sign_jwt
from data_rentgen.services import UnitOfWork

logger = logging.getLogger(__name__)


class DummyAuthProvider(AuthProvider):
    def __init__(
        self,
        settings: Annotated[DummyAuthProviderSettings, Depends(Stub(DummyAuthProviderSettings))],
        unit_of_work: Annotated[UnitOfWork, Depends()],
    ) -> None:
        self._settings = settings
        self._uow = unit_of_work

    @classmethod
    def setup(cls, app: FastAPI) -> FastAPI:
        settings = DummyAuthProviderSettings.model_validate(
            app.state.settings.auth.model_dump(exclude={"provider"}),
        )
        logger.info("Using %s provider with settings:\n%s", cls.__name__, pformat(settings))
        app.dependency_overrides[AuthProvider] = cls
        app.dependency_overrides[DummyAuthProviderSettings] = lambda: settings
        return app

    async def get_current_user(self, access_token: str | None, request: Request) -> User:
        if not access_token:
            msg = "Missing auth credentials"
            raise AuthorizationError(msg)

        user_id = self._get_user_id_from_token(access_token)
        user = await self._uow.user.read_by_id(user_id)
        if user is None:
            raise EntityNotFoundError("User", "user_id", user_id)  # type: ignore[call-arg]  # noqa: EM101
        return user

    async def get_token_password_grant(
        self,
        login: str,
        password: str,
    ) -> dict[str, Any]:
        if not login:
            msg = "Missing auth credentials"
            raise AuthorizationError(msg)

        logger.debug("Get/create user %r in database", login)
        async with self._uow:
            user = await self._uow.user.get_or_create(UserDTO(name=login))

        logger.debug("User with id %r found", user.id)
        logger.debug("Generate access token for user id %r", user.id)
        access_token, expires_at = self._generate_access_token(user_id=user.id)
        return {
            "access_token": access_token,
            "token_type": "bearer",
            "expires_at": expires_at,
        }

    async def get_token_authorization_code_grant(
        self,
        code: str,
    ) -> dict[str, Any]:
        msg = "Authorization code grant is not supported by DummyAuthProvider."
        raise NotImplementedError(msg)

    async def logout(self, user: User, refresh_token: str | None) -> None:
        msg = "Logout method is not implemented for DummyAuthProvider."
        raise NotImplementedError(msg)

    def _generate_access_token(self, user_id: int) -> tuple[str, float]:
        expires_at = time() + self._settings.access_token.expire_seconds
        payload = {
            "user_id": user_id,
            "exp": expires_at,
        }
        access_token = sign_jwt(
            payload,
            self._settings.access_token.secret_key.get_secret_value(),
            self._settings.access_token.security_algorithm,
        )
        return access_token, expires_at

    def _get_user_id_from_token(self, token: str) -> int:
        try:
            payload = decode_jwt(
                token,
                self._settings.access_token.secret_key.get_secret_value(),
                self._settings.access_token.security_algorithm,
            )
            return int(payload["user_id"])
        except (KeyError, TypeError, ValueError) as e:
            msg = "Invalid token"
            raise AuthorizationError(msg) from e



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import secrets
import time
from datetime import datetime, timezone
from hashlib import sha1
from uuid import NAMESPACE_URL, uuid5
from uuid import UUID as BaseUUID  # noqa: N811

from uuid6 import UUID as NewUUID  # noqa: N811

__all__ = ["extract_timestamp_from_uuid", "generate_incremental_uuid", "generate_new_uuid", "generate_static_uuid"]


def generate_new_uuid(instant: datetime | None = None) -> NewUUID:
    """Generate new UUID for an instant of time. Each function call returns a new UUID value.

    UUID version is an implementation detail, and **should not** be relied on.
    For now it is `UUIDv7 <https://datatracker.ietf.org/doc/rfc9562/>`_, so for increasing instant values,
    returned UUID is always greater than previous one.

    Using uuid6 lib implementation (MIT License), with few changes:
    * https://github.com/oittaa/uuid6-python/blob/4f879849178b8a7a564f7cb76c3f7a6e5228d9ed/src/uuid6/__init__.py#L128-L147
    * https://github.com/oittaa/uuid6-python/blob/4f879849178b8a7a564f7cb76c3f7a6e5228d9ed/src/uuid6/__init__.py#L46-L51
    """

    timestamp_ms = int(instant.timestamp() * 1000) if instant else time.time_ns() // 10**6
    node = secrets.randbits(76)
    return _build_uuidv7(timestamp_ms, node)


def generate_incremental_uuid(instant: datetime, data: str) -> NewUUID:
    """Generate new UUID for an instant of time and data. Each function call with the same arguments returns the same result.

    UUID version is an implementation detail, and **should not** be relied on.
    For now it is `UUIDv7 <https://datatracker.ietf.org/doc/rfc9562/>`_, so for increasing instant values,
    returned UUID is always greater than previous one.

    Using uuid6 lib implementation (MIT License), with few changes:
    * https://github.com/oittaa/uuid6-python/blob/4f879849178b8a7a564f7cb76c3f7a6e5228d9ed/src/uuid6/__init__.py#L128-L147
    * https://github.com/oittaa/uuid6-python/blob/4f879849178b8a7a564f7cb76c3f7a6e5228d9ed/src/uuid6/__init__.py#L46-L51
    """  # noqa: E501

    instant_utc = instant.astimezone(timezone.utc)
    timestamp_ms = int(instant_utc.timestamp() * 1000)

    # generate the rest of bytes using some hash.
    # can be used to generate consistent UUIDs for some input, e.g. external runId.
    # if data is some static value, e.g. job name, mix it with timestamp to make it more random
    digest = sha1(instant_utc.isoformat().encode("utf-8") + data.encode("utf-8"), usedforsecurity=False).digest()
    # sha1 returns 160bit hash, we need only first 76 bits
    node = int(digest.hex(), 16) >> 84

    return _build_uuidv7(timestamp_ms, node)


def _build_uuidv7(timestamp: int, node: int) -> NewUUID:
    # merge timestamp and node into 128-bit UUID
    # timestamp is first 48 bits, node is last 80 bits
    uuid_int = (timestamp & 0xFFFFFFFFFFFF) << 80
    uuid_int |= node & 0xFFFFFFFFFFFFFFFFFFFF

    # Set the version number (4 bit).
    version = 7
    uuid_int &= ~(0xF000 << 64)
    uuid_int |= version << 76

    # Set the variant (2 bit) to RFC 4122.
    uuid_int &= ~(0xC000 << 48)
    uuid_int |= 0x8000 << 48

    return NewUUID(int=uuid_int)


def generate_static_uuid(data: str) -> BaseUUID:
    """Generate static UUID for data. Each function call returns the same UUID value.

    UUID version is an implementation detailed, and **should not** be relied on.
    For now it is UUIDv5 with namespace=URL.
    """
    return uuid5(namespace=NAMESPACE_URL, name=data)  # type: ignore[arg-type]


def extract_timestamp_from_uuid(uuid: BaseUUID) -> datetime:
    """Extract timestamp from UUIDv7"""
    uuid = NewUUID(int=uuid.int)
    if not uuid.version or uuid.version < 6:  # noqa: PLR2004
        msg = "Only UUIDv6+ are supported"
        raise ValueError(msg)
    return datetime.fromtimestamp(uuid.time / 1000, tz=timezone.utc)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends

from data_rentgen.db.models.user import User
from data_rentgen.server.errors.registration import get_error_responses
from data_rentgen.server.schemas.v1.user import UserResponseV1
from data_rentgen.server.services import get_user

router = APIRouter(
    prefix="/users",
    tags=["User"],
    responses=get_error_responses(),
)


@router.get("/me")
async def check_auth(
    current_user: Annotated[User, Depends(get_user())],
) -> UserResponseV1:
    return UserResponseV1.model_validate(current_user)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create run

Revision ID: 5f8fff06dd76
Revises: 026de1556610
Create Date: 2024-06-27 19:14:50.909604

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "5f8fff06dd76"
down_revision = "026de1556610"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "run",
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("job_id", sa.BigInteger(), nullable=False),
        sa.Column("parent_run_id", sa.UUID(), nullable=True),
        sa.Column("status", sa.SmallInteger(), nullable=False),
        sa.Column("external_id", sa.String(), nullable=True),
        sa.Column("attempt", sa.String(length=64), nullable=True),
        sa.Column("persistent_log_url", sa.String(), nullable=True),
        sa.Column("running_log_url", sa.String(), nullable=True),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("started_by_user_id", sa.BigInteger(), nullable=True),
        sa.Column("start_reason", sa.String(length=32), nullable=True),
        sa.Column("ended_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("end_reason", sa.String(), nullable=True),
        sa.Column(
            "search_vector",
            postgresql.TSVECTOR(),
            sa.Computed(
                "to_tsvector('simple'::regconfig, COALESCE(external_id, ''::text) || ' ' || (translate(external_id, '/.', '  ')))",
                persisted=True,
            ),
        ),
        sa.PrimaryKeyConstraint("created_at", "id", name=op.f("pk__run")),
        postgresql_partition_by="RANGE (created_at)",
    )
    op.create_index(op.f("ix__run__job_id"), "run", ["job_id"], unique=False)
    op.create_index(op.f("ix__run__parent_run_id"), "run", ["parent_run_id"], unique=False)
    op.create_index(op.f("ix__run__search_vector"), "run", ["search_vector"], unique=False, postgresql_using="gin")


def downgrade() -> None:
    op.drop_index(op.f("ix__run__parent_run_id"), table_name="run")
    op.drop_index(op.f("ix__run__job_id"), table_name="run")
    op.drop_index(op.f("ix__run__search_vector"), table_name="run", postgresql_using="gin")
    op.drop_table("run")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)


class OpenLineageColumnLineageDatasetFacetFieldTransformation(OpenLineageBase):
    """Dataset facet describing field transformation."""

    type: str
    subtype: str | None = None
    description: str | None = None
    masking: bool = False


class OpenLineageColumnLineageDatasetFacetFieldRef(OpenLineageBase):
    """Dataset facet describing field reference for column lineage facet."""

    namespace: str
    name: str
    field: str
    transformations: list[OpenLineageColumnLineageDatasetFacetFieldTransformation]


class OpenLineageColumnLineageDatasetFacetField(OpenLineageBase):
    """Dataset facet describing column lineage for specific field."""

    inputFields: list[OpenLineageColumnLineageDatasetFacetFieldRef] = Field(default_factory=list)
    transformationDescription: str | None = None
    transformationType: str | None = None


class OpenLineageColumnLineageDatasetFacet(OpenLineageDatasetFacet):
    """Dataset facet describing column lineage.
    See [InputStatisticsInputDatasetFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/ColumnLineageDatasetFacet.json).
    """

    fields: dict[str, OpenLineageColumnLineageDatasetFacetField] = Field(default_factory=dict)
    dataset: list[OpenLineageColumnLineageDatasetFacetFieldRef] = Field(default_factory=list)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import logging
from contextlib import asynccontextmanager
from typing import Any

import anyio
from fast_depends import dependency_provider
from faststream import ContextRepo, FastStream
from faststream._compat import ExceptionGroup
from faststream.asgi import AsgiFastStream, AsgiResponse, get
from faststream.kafka import KafkaBroker
from sqlalchemy.ext.asyncio import AsyncSession

import data_rentgen
from data_rentgen.consumer.settings import ConsumerApplicationSettings
from data_rentgen.consumer.subscribers import runs_events_subscriber
from data_rentgen.db.factory import create_session_factory
from data_rentgen.logging.setup_logging import setup_logging

logger = logging.getLogger(__name__)


@get  # type: ignore[arg-type]
async def liveness(scope: dict[str, Any]) -> AsgiResponse:
    return AsgiResponse(b"", status_code=204)


def broker_factory(settings: ConsumerApplicationSettings) -> KafkaBroker:
    broker = KafkaBroker(
        bootstrap_servers=settings.kafka.bootstrap_servers,
        security=settings.kafka.security.to_security(),
        compression_type=settings.kafka.compression.value if settings.kafka.compression else None,
        client_id=f"data-rentgen-{data_rentgen.__version__}",
        logger=logger,
        **settings.kafka.security.extra_broker_kwargs(),
    )

    # register subscribers using settings
    consumer_settings = settings.consumer.model_dump(exclude={"topics_list", "topics_pattern"})
    broker.subscriber(
        *settings.consumer.topics_list,
        pattern=settings.consumer.topics_pattern,
        **consumer_settings,
        batch=True,
    )(runs_events_subscriber)

    dependency_provider.override(AsyncSession, create_session_factory(settings.database))
    return broker


def application_factory(settings: ConsumerApplicationSettings) -> AsgiFastStream:
    @asynccontextmanager
    async def security_lifespan(context: ContextRepo):
        try:
            async with anyio.create_task_group() as tg:
                await settings.kafka.security.initialize()
                tg.start_soon(settings.kafka.security.refresh)

                yield

                await settings.kafka.security.destroy()
                tg.cancel_scope.cancel()
        except ExceptionGroup as e:
            for exception in e.exceptions:
                raise exception from None

    return FastStream(
        broker=broker_factory(settings),
        lifespan=security_lifespan,
        title="Data.Rentgen",
        description="Data.Rentgen is a nextgen DataLineage service",
        version=data_rentgen.__version__,
        logger=logger,
    ).as_asgi(asgi_routes=[("/internal/healthcheck/liveness", liveness)])


def get_application():
    settings = ConsumerApplicationSettings()
    setup_logging(settings.logging)
    return application_factory(settings=settings)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.dataset_facets.base import (
    OpenLineageDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.column_lineage import (
    OpenLineageColumnLineageDatasetFacet,
    OpenLineageColumnLineageDatasetFacetField,
    OpenLineageColumnLineageDatasetFacetFieldRef,
    OpenLineageColumnLineageDatasetFacetFieldTransformation,
)
from data_rentgen.consumer.openlineage.dataset_facets.datasource import (
    OpenLineageDatasourceDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.documentation import (
    OpenLineageDocumentationDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.input_statistics import (
    OpenLineageInputStatisticsInputDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.lifecycle_change import (
    OpenLineageDatasetLifecycleStateChange,
    OpenLineageDatasetPreviousIdentifier,
    OpenLineageLifecycleStateChangeDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.output_statistics import (
    OpenLineageOutputStatisticsOutputDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.schema import (
    OpenLineageSchemaDatasetFacet,
    OpenLineageSchemaField,
)
from data_rentgen.consumer.openlineage.dataset_facets.storage import (
    OpenLineageStorageDatasetFacet,
)
from data_rentgen.consumer.openlineage.dataset_facets.symlinks import (
    OpenLineageSymlinkIdentifier,
    OpenLineageSymlinksDatasetFacet,
    OpenLineageSymlinkType,
)

__all__ = [
    "OpenLineageColumnLineageDatasetFacet",
    "OpenLineageColumnLineageDatasetFacetField",
    "OpenLineageColumnLineageDatasetFacetFieldRef",
    "OpenLineageColumnLineageDatasetFacetFieldTransformation",
    "OpenLineageDatasetFacet",
    "OpenLineageDatasetFacets",
    "OpenLineageDatasetLifecycleStateChange",
    "OpenLineageDatasetPreviousIdentifier",
    "OpenLineageDatasourceDatasetFacet",
    "OpenLineageDocumentationDatasetFacet",
    "OpenLineageInputDatasetFacets",
    "OpenLineageInputStatisticsInputDatasetFacet",
    "OpenLineageLifecycleStateChangeDatasetFacet",
    "OpenLineageOutputDatasetFacets",
    "OpenLineageOutputStatisticsOutputDatasetFacet",
    "OpenLineageSchemaDatasetFacet",
    "OpenLineageSchemaField",
    "OpenLineageStorageDatasetFacet",
    "OpenLineageSymlinkIdentifier",
    "OpenLineageSymlinkType",
    "OpenLineageSymlinksDatasetFacet",
]


class OpenLineageDatasetFacets(OpenLineageBase):
    """All possible dataset facets.
    See [Dataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    documentation: OpenLineageDocumentationDatasetFacet | None = None
    dataSource: OpenLineageDatasourceDatasetFacet | None = None
    lifecycleStateChange: OpenLineageLifecycleStateChangeDatasetFacet | None = None
    datasetSchema: OpenLineageSchemaDatasetFacet | None = Field(default=None, alias="schema")
    storage: OpenLineageStorageDatasetFacet | None = None
    symlinks: OpenLineageSymlinksDatasetFacet | None = None
    columnLineage: OpenLineageColumnLineageDatasetFacet | None = None


class OpenLineageInputDatasetFacets(OpenLineageBase):
    """All possible input dataset facets.
    See [InputDataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    inputStatistics: OpenLineageInputStatisticsInputDatasetFacet | None = None


class OpenLineageOutputDatasetFacets(OpenLineageBase):
    """All possible output dataset facets.
    See [InputDataset](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    outputStatistics: OpenLineageOutputStatisticsOutputDatasetFacet | None = None



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from dataclasses import dataclass, field


@dataclass
class LocationDTO:
    type: str
    name: str
    addresses: set[str]
    id: int | None = field(default=None, compare=False)

    @property
    def unique_key(self) -> tuple:
        return (self.type, self.name)

    def merge(self, new: LocationDTO) -> LocationDTO:
        if new.id is None and new.addresses.issubset(self.addresses):
            # locations aren't changed that much, reuse them if possible
            return self

        return LocationDTO(
            type=self.type,
            name=self.name,
            addresses=self.addresses | new.addresses,
            id=new.id or self.id,
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import http
from typing import Any, Literal

from data_rentgen.exceptions.auth import AuthorizationError
from data_rentgen.exceptions.redirect import RedirectError
from data_rentgen.server.errors.base import BaseErrorSchema
from data_rentgen.server.errors.registration import register_error_response


@register_error_response(
    exception=AuthorizationError,
    status=http.HTTPStatus.UNAUTHORIZED,
)
class NotAuthorizedSchema(BaseErrorSchema):
    code: Literal["unauthorized"] = "unauthorized"
    details: Any = None


@register_error_response(
    exception=RedirectError,
    status=http.HTTPStatus.UNAUTHORIZED,
)
class NotAuthorizedRedirectSchema(BaseErrorSchema):
    """
    The reason of using UNAUTHORIZED instead of strict redirect is:
    Fronted is using `fetch()` function which can't handle redirect responses
    https://github.com/whatwg/fetch/issues/601
    """

    code: Literal["auth_redirect"] = "auth_redirect"



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from fastapi import FastAPI

from data_rentgen.server.middlewares.application_version import (
    apply_application_version_middleware,
)
from data_rentgen.server.middlewares.cors import apply_cors_middleware
from data_rentgen.server.middlewares.monitoring import (
    apply_monitoring_metrics_middleware,
)
from data_rentgen.server.middlewares.openapi import apply_openapi_middleware
from data_rentgen.server.middlewares.request_id import apply_request_id_middleware
from data_rentgen.server.middlewares.session import apply_session_middleware
from data_rentgen.server.middlewares.static_files import apply_static_files
from data_rentgen.server.settings import ServerSettings


def apply_middlewares(
    application: FastAPI,
    settings: ServerSettings,
) -> FastAPI:
    """Add middlewares to the application."""

    apply_cors_middleware(application, settings.cors)
    apply_monitoring_metrics_middleware(application, settings.monitoring)
    apply_request_id_middleware(application, settings.request_id)
    apply_application_version_middleware(application, settings.application_version)
    apply_openapi_middleware(application, settings.openapi)
    apply_static_files(application, settings.static_files)
    apply_session_middleware(application, settings.session)

    return application



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import jwt

from data_rentgen.exceptions.auth import AuthorizationError


def sign_jwt(payload: dict, secret_key: str, security_algorithm: str) -> str:
    return jwt.encode(
        payload=payload,
        key=secret_key,
        algorithm=security_algorithm,
    )


def decode_jwt(token: str, secret_key: str, security_algorithm: str) -> dict:
    try:
        claims = jwt.decode(jwt=token, key=secret_key, algorithms=[security_algorithm])

        if "exp" not in claims:
            err_msg = "Missing expiration time in token"
            raise jwt.ExpiredSignatureError(err_msg)

    except jwt.PyJWTError as e:
        err_msg = "Invalid token"
        raise AuthorizationError(err_msg) from e

    else:
        return claims



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Collection
from uuid import UUID

from sqlalchemy import any_, select

from data_rentgen.db.models import Schema
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.fields import get_fields_digest
from data_rentgen.dto import SchemaDTO


class SchemaRepository(Repository[Schema]):
    async def get_or_create(self, schema: SchemaDTO) -> Schema:
        # avoid calculating digest twice
        digest = get_fields_digest(schema.fields)
        result = await self._get(digest)
        if not result:
            # try one more time, but with lock acquired.
            # if another worker already created the same row, just use it. if not - create with holding the lock.
            await self._lock(digest)
            result = await self._get(digest) or await self._create(digest, schema)
        return result

    async def list_by_ids(self, schema_ids: Collection[int]) -> list[Schema]:
        if not schema_ids:
            return []

        query = select(Schema).where(Schema.id == any_(list(schema_ids)))  # type: ignore[arg-type]
        result = await self._session.scalars(query)
        return list(result.all())

    async def _get(self, digest: UUID) -> Schema | None:
        result = select(Schema).where(Schema.digest == digest)
        return await self._session.scalar(result)

    async def _create(self, digest: UUID, schema: SchemaDTO) -> Schema:
        result = Schema(digest=digest, fields=schema.fields)
        self._session.add(result)
        await self._session.flush([result])
        return result



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from fastapi import APIRouter

from data_rentgen.server.api.v1.router.auth import router as auth_router
from data_rentgen.server.api.v1.router.dataset import router as dataset_router
from data_rentgen.server.api.v1.router.job import router as job_router
from data_rentgen.server.api.v1.router.location import router as location_router
from data_rentgen.server.api.v1.router.operation import router as operation_router
from data_rentgen.server.api.v1.router.run import router as run_router
from data_rentgen.server.api.v1.router.user import router as user_router

router = APIRouter(prefix="/v1")
router.include_router(auth_router)
router.include_router(dataset_router)
router.include_router(job_router)
router.include_router(location_router)
router.include_router(operation_router)
router.include_router(run_router)
router.include_router(user_router)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.consumer.openlineage.job_facets.base import OpenLineageJobFacet


class OpenLineageDocumentationJobFacet(OpenLineageJobFacet):
    """Job facet describing documentation.
    See [DocumentationJobFacet](https://github.com/OpenLineage/OpenLineage/blob/main/spec/facets/DocumentationJobFacet.json).
    """

    description: str



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

from pydantic import BaseModel, Field, field_validator


class StaticFilesSettings(BaseModel):
    """Static files serving settings.

    Files are served at ``/static`` endpoint.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__SERVER__STATIC_FILES__ENABLED=True
        DATA_RENTGEN__SERVER__STATIC_FILES__DIRECTORY=/app/data_rentgen/server/static
    """

    enabled: bool = Field(default=True, description="Set to ``True`` to enable static file serving")
    directory: Path = Field(
        default=Path("docs/_static"),
        description="Directory containing static files",
    )

    @field_validator("directory")
    def _validate_directory(cls, value: Path) -> Path:  # noqa: N805
        if not value.exists():
            msg = f"Directory '{value}' does not exist"
            raise ValueError(msg)
        if not value.is_dir():
            msg = f"Path '{value}' is not a directory"
            raise ValueError(msg)
        return value



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from logging.config import dictConfig
from pathlib import Path

import yaml

from data_rentgen.logging.exceptions import LoggingSetupError
from data_rentgen.logging.settings import LoggingSettings

PRESETS_PATH = Path(__file__).parent.joinpath("presets").resolve()


def setup_logging(settings: LoggingSettings) -> None:
    """Parse file with logging configuration, and setup logging accordingly"""
    if not settings.setup:
        return

    config_path = settings.custom_config_path or PRESETS_PATH.joinpath(f"{settings.preset}.yml")

    if not config_path.exists():
        msg = f"Logging configuration file '{config_path}' does not exist"
        raise OSError(msg)

    try:
        config = yaml.safe_load(config_path.read_text())
        dictConfig(config)
    except Exception as e:
        msg = f"Error reading logging configuration '{config_path}'"
        raise LoggingSetupError(msg) from e



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import asyncio
import logging
import os
import subprocess
from contextlib import suppress
from pathlib import Path
from typing import Any, Literal

import anyio
from faststream.security import SASLGSSAPI
from pydantic import Field, FilePath, SecretStr, model_validator

from data_rentgen.consumer.settings.security.base import KafkaSecurityBaseSettings

logger = logging.getLogger(__name__)


class KafkaSecurityGSSAPISettings(KafkaSecurityBaseSettings):
    """Kafka GSSAPI auth settings.

    This auth method requires installing ``data-rentgen[gssapi]`` extra,
    and relies on presence of ``kinit`` and ``kdestroy`` binaries in your operating system.

    Examples
    --------

    Using principal + password for calling ``kinit``:

    .. code-block:: bash

        DATA_RENTGEN__KAFKA__SECURITY__TYPE=GSSAPI
        DATA_RENTGEN__KAFKA__SECURITY__PRINCIPAL=dummy
        DATA_RENTGEN__KAFKA__SECURITY__PASSWORD=changeme
        DATA_RENTGEN__KAFKA__SECURITY__REALM=MY.REALM.COM

    Using principal + keytab for calling ``kinit``:

    .. code-block:: bash

        DATA_RENTGEN__KAFKA__SECURITY__TYPE=GSSAPI
        DATA_RENTGEN__KAFKA__SECURITY__PRINCIPAL=dummy
        DATA_RENTGEN__KAFKA__SECURITY__KEYTAB=/etc/security/dummy.keytab
        DATA_RENTGEN__KAFKA__SECURITY__REALM=MY.REALM.COM
    """

    type: Literal["GSSAPI"] = "GSSAPI"
    principal: str = Field(description="Kerberos principal")
    password: SecretStr | None = Field(description="Kerberos password. Mutually exclusive with keytab")
    keytab: FilePath | None = Field(default=None, description="Path to keytab file. Mutually exclusive with password")
    realm: str | None = Field(default=None, description="Kerberos realm")
    service_name: str = Field(default="kafka", description="Kerberos service name")
    refresh_interval_seconds: int = Field(default=60 * 60, description="Refresh Kerberos ticket every N seconds")

    @model_validator(mode="after")
    def _check_password_or_keytab(self):
        if not (self.password or self.keytab):
            msg = "input should contain either 'password' or 'keytab' field, both are empty"
            raise ValueError(msg)
        if self.password and self.keytab:
            msg = "input should contain either 'password' or 'keytab' field, both are set"
            raise ValueError(msg)
        return self

    def to_security(self):
        return SASLGSSAPI()

    def extra_broker_kwargs(self) -> dict[str, Any]:
        return {
            "sasl_kerberos_service_name": self.service_name,
            "sasl_kerberos_domain_name": self.realm,
        }

    async def kinit_password(self):
        cmd = ["kinit", self.principal]
        logger.debug("Calling command: %s", " ".join(cmd))

        process = await asyncio.create_subprocess_exec(
            *cmd,
            stdin=asyncio.subprocess.PIPE,
            # do not show user 'Please enter password' banner
            stdout=asyncio.subprocess.PIPE,
            # do not capture stderr, immediately show all errors to user
        )
        password: str = self.password.get_secret_value()  # type: ignore[assignment]
        stdout, stderr = await process.communicate(password.encode("utf-8"))
        retcode = await process.wait()
        if retcode:
            raise subprocess.CalledProcessError(retcode, cmd, stdout, stderr)

    async def kinit_keytab(self):
        keytab: Path = self.keytab.resolve()  # type: ignore[assignment]
        cmd = ["kinit", self.principal, "-k", "-t", os.fspath(keytab)]
        logger.debug("Calling command: %s", " ".join(cmd))

        process = await asyncio.create_subprocess_exec(*cmd)
        retcode = await process.wait()
        if retcode:
            raise subprocess.CalledProcessError(retcode, cmd)

    async def kinit_refresh(self):
        cmd = ["kinit", "-R", self.principal]
        logger.debug("Calling command: %s", " ".join(cmd))

        process = await asyncio.create_subprocess_exec(*cmd)
        retcode = await process.wait()
        if retcode:
            raise subprocess.CalledProcessError(retcode, cmd)

    async def kdestroy(self):
        cmd = ["kdestroy", "-p", self.principal]
        logger.debug("Calling command: %s", " ".join(cmd))

        process = await asyncio.create_subprocess_exec(*cmd)
        retcode = await process.wait()
        if retcode:
            raise subprocess.CalledProcessError(retcode, cmd)

    async def initialize(self) -> None:
        if self.keytab:
            await self.kinit_keytab()
        elif self.password:
            await self.kinit_password()

    async def refresh(self) -> None:
        while True:
            await anyio.sleep(self.refresh_interval_seconds)

            with suppress(Exception):
                # if ticket is renewable, try to refresh it
                await self.kinit_refresh()
                continue

            await self.initialize()

    async def destroy(self) -> None:
        with suppress(Exception):
            await self.kdestroy()



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create column lineage

Revision ID: 15c0a22b8566
Revises: f017d4c58658
Create Date: 2025-02-04 11:30:21.190161

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "15c0a22b8566"
down_revision = "f017d4c58658"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "dataset_column_relation",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("fingerprint", sa.UUID(), nullable=False),
        sa.Column("source_column", sa.String(length=255), nullable=False),
        sa.Column("target_column", sa.String(length=255), nullable=True),
        sa.Column("type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__dataset_column_relation")),
    )
    op.create_index(
        op.f("ix__dataset_column_relation__fingerprint_source_column_target_column"),
        "dataset_column_relation",
        ["fingerprint", "source_column", sa.text("coalesce(target_column, '')")],
        unique=True,
    )

    op.create_table(
        "column_lineage",
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("operation_id", sa.UUID(), nullable=False),
        sa.Column("run_id", sa.UUID(), nullable=False),
        sa.Column("job_id", sa.BigInteger(), nullable=False),
        sa.Column("source_dataset_id", sa.BigInteger(), nullable=False),
        sa.Column("target_dataset_id", sa.BigInteger(), nullable=False),
        sa.Column("fingerprint", sa.UUID(), nullable=False),
        sa.PrimaryKeyConstraint("created_at", "id", name=op.f("pk__column_lineage")),
        postgresql_partition_by="RANGE (created_at)",
    )
    op.create_index(op.f("ix__column_lineage__job_id"), "column_lineage", ["job_id"], unique=False)
    op.create_index(op.f("ix__column_lineage__operation_id"), "column_lineage", ["operation_id"], unique=False)
    op.create_index(op.f("ix__column_lineage__run_id"), "column_lineage", ["run_id"], unique=False)
    op.create_index(
        op.f("ix__column_lineage__source_dataset_id"),
        "column_lineage",
        ["source_dataset_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix__column_lineage__target_dataset_id"),
        "column_lineage",
        ["target_dataset_id"],
        unique=False,
    )


def downgrade() -> None:
    op.drop_index(op.f("ix__column_lineage__target_dataset_id"), table_name="column_lineage")
    op.drop_index(op.f("ix__column_lineage__source_dataset_id"), table_name="column_lineage")
    op.drop_index(op.f("ix__column_lineage__run_id"), table_name="column_lineage")
    op.drop_index(op.f("ix__column_lineage__operation_id"), table_name="column_lineage")
    op.drop_index(op.f("ix__column_lineage__job_id"), table_name="column_lineage")
    op.drop_table("column_lineage")

    op.drop_index(
        op.f("ix__dataset_column_relation__fingerprint_source_column_target_column"),
        table_name="dataset_column_relation",
    )
    op.drop_table("dataset_column_relation")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create output

Revision ID: 3357c8f914aa
Revises: 412d5fbdb362
Create Date: 2024-06-27 19:25:51.909604

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3357c8f914aa"
down_revision = "412d5fbdb362"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "output",
        sa.Column("created_at", sa.DateTime(timezone=True), nullable=False),
        sa.Column("id", sa.UUID(), nullable=False),
        sa.Column("operation_id", sa.UUID(), nullable=False),
        sa.Column("dataset_id", sa.BigInteger(), nullable=False),
        sa.Column("run_id", sa.UUID(), nullable=False),
        sa.Column("job_id", sa.BigInteger(), nullable=False),
        sa.Column("type", sa.String(length=32), nullable=False),
        sa.Column("schema_id", sa.BigInteger(), nullable=True),
        sa.Column("num_bytes", sa.BigInteger(), nullable=True),
        sa.Column("num_rows", sa.BigInteger(), nullable=True),
        sa.Column("num_files", sa.BigInteger(), nullable=True),
        sa.PrimaryKeyConstraint("created_at", "id", name=op.f("pk__output")),
        postgresql_partition_by="RANGE (created_at)",
    )
    op.create_index(op.f("ix__output__dataset_id"), "output", ["dataset_id"], unique=False)
    op.create_index(op.f("ix__output__operation_id"), "output", ["operation_id"], unique=False)
    op.create_index(op.f("ix__output__run_id"), "output", ["run_id"], unique=False)
    op.create_index(op.f("ix__output__job_id"), "output", ["job_id"], unique=False)
    op.create_index(op.f("ix__output__schema_id"), "output", ["schema_id"], unique=False)


def downgrade() -> None:
    op.drop_index(op.f("ix__output__schema_id"), table_name="output")
    op.drop_index(op.f("ix__output__job_id"), table_name="output")
    op.drop_index(op.f("ix__output__run_id"), table_name="output")
    op.drop_index(op.f("ix__output__operation_id"), table_name="output")
    op.drop_index(op.f("ix__output__dataset_id"), table_name="output")
    op.drop_table("output")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create address

Revision ID: c324e48e0f71
Revises: e730f34b9893
Create Date: 2024-06-27 18:44:50.259710

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c324e48e0f71"
down_revision = "e730f34b9893"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "address",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("location_id", sa.BigInteger(), nullable=False),
        sa.Column("url", sa.String(), nullable=False),
        sa.Column(
            "search_vector",
            postgresql.TSVECTOR(),
            sa.Computed(
                "to_tsvector('simple'::regconfig, url || ' ' || (translate(url, '/.', '  ')))",
                persisted=True,
            ),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["location_id"],
            ["location.id"],
            name=op.f("fk__address__location_id__location"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__address")),
        sa.UniqueConstraint("location_id", "url", name=op.f("uq__address__location_id_url")),
    )
    op.create_index(op.f("ix__address__location_id"), "address", ["location_id"], unique=False)
    op.create_index(op.f("ix__address__url"), "address", ["url"], unique=False)
    op.create_index(
        op.f("ix__address__search_vector"),
        "address",
        ["search_vector"],
        unique=False,
        postgresql_using="gin",
    )


def downgrade() -> None:
    op.drop_index(op.f("ix__address__url"), table_name="address")
    op.drop_index(op.f("ix__address__location_id"), table_name="address")
    op.drop_index(op.f("ix__address__search_vector"), table_name="address", postgresql_using="gin")
    op.drop_table("address")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection

from sqlalchemy import (
    ColumnElement,
    CompoundSelect,
    Row,
    Select,
    SQLColumnExpression,
    any_,
    asc,
    desc,
    func,
    select,
    union,
)
from sqlalchemy.orm import selectinload

from data_rentgen.db.models import Address, Job, JobType, Location
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.search import make_tsquery, ts_match, ts_rank
from data_rentgen.dto import JobDTO, PaginationDTO


class JobRepository(Repository[Job]):
    async def paginate(
        self,
        page: int,
        page_size: int,
        job_ids: Collection[int],
        search_query: str | None,
    ) -> PaginationDTO[Job]:
        where = []
        if job_ids:
            where.append(Job.id == any_(list(job_ids)))  # type: ignore[arg-type]

        query: Select | CompoundSelect
        order_by: list[ColumnElement | SQLColumnExpression]
        if search_query:
            tsquery = make_tsquery(search_query)

            job_stmt = select(Job, ts_rank(Job.search_vector, tsquery).label("search_rank")).where(
                ts_match(Job.search_vector, tsquery),
                *where,
            )
            location_stmt = (
                select(Job, ts_rank(Location.search_vector, tsquery).label("search_rank"))
                .join(Job, Location.id == Job.location_id)
                .where(ts_match(Location.search_vector, tsquery), *where)
            )
            address_stmt = (
                select(Job, func.max(ts_rank(Address.search_vector, tsquery).label("search_rank")))
                .join(Location, Address.location_id == Location.id)
                .join(Job, Location.id == Job.location_id)
                .where(ts_match(Address.search_vector, tsquery), *where)
                .group_by(Job.id, Location.id, Address.id)
            )

            union_cte = union(job_stmt, location_stmt, address_stmt).cte()

            job_columns = [column for column in union_cte.columns if column.name != "search_rank"]

            query = select(
                *job_columns,
                func.max(union_cte.c.search_rank).label("search_rank"),
            ).group_by(*job_columns)
            order_by = [desc("search_rank"), asc("name")]
        else:
            query = select(Job).where(*where)
            order_by = [Job.name]

        options = [selectinload(Job.location).selectinload(Location.addresses)]
        return await self._paginate_by_query(
            query=query,
            order_by=order_by,
            options=options,
            page=page,
            page_size=page_size,
        )

    async def create_or_update(self, job: JobDTO) -> Job:
        result = await self._get(job)
        if not result:
            # try one more time, but with lock acquired.
            # if another worker already created the same row, just use it. if not - create with holding the lock.
            await self._lock(job.location.id, job.name)
            result = await self._get(job)

        if not result:
            return await self._create(job)
        return await self._update(result, job)

    async def list_by_ids(self, job_ids: Collection[int]) -> list[Job]:
        if not job_ids:
            return []
        query = (
            select(Job)
            .where(Job.id == any_(list(job_ids)))  # type: ignore[arg-type]
            .options(selectinload(Job.location).selectinload(Location.addresses))
        )
        result = await self._session.scalars(query)
        return list(result.all())

    async def get_stats_by_location_ids(self, location_ids: Collection[int]) -> dict[int, Row]:
        if not location_ids:
            return {}

        query = (
            select(
                Job.location_id.label("location_id"),
                func.count(Job.id.distinct()).label("total_jobs"),
            )
            .where(
                Job.location_id == any_(list(location_ids)),  # type: ignore[arg-type]
            )
            .group_by(Job.location_id)
        )

        query_result = await self._session.execute(query)
        return {row.location_id: row for row in query_result.all()}

    async def _get(self, job: JobDTO) -> Job | None:
        statement = select(Job).where(Job.location_id == job.location.id, Job.name == job.name)
        return await self._session.scalar(statement)

    async def _create(self, job: JobDTO) -> Job:
        result = Job(
            location_id=job.location.id,
            name=job.name,
            type=JobType(job.type) if job.type else JobType.UNKNOWN,
        )
        self._session.add(result)
        await self._session.flush([result])
        return result

    async def _update(self, existing: Job, new: JobDTO) -> Job:
        # almost of fields are immutable, so we can avoid UPDATE statements if row is unchanged
        if new.type:
            existing.type = JobType(new.type)
            await self._session.flush([existing])
        return existing



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from pydantic import Field

from data_rentgen.consumer.openlineage.base import OpenLineageBase
from data_rentgen.consumer.openlineage.run_facets import OpenLineageRunFacets
from data_rentgen.utils import UUIDv6Plus


class OpenLineageRun(OpenLineageBase):
    """Run model.
    See [Run](https://github.com/OpenLineage/OpenLineage/blob/main/spec/OpenLineage.json).
    """

    runId: UUIDv6Plus
    facets: OpenLineageRunFacets = Field(default_factory=OpenLineageRunFacets)



# SPDX-FileCopyrightText: 2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from sqlalchemy import BigInteger, ForeignKey, text
from sqlalchemy.dialects.postgresql import JSONB, TSTZRANGE, ExcludeConstraint
from sqlalchemy.orm import Mapped, mapped_column

from data_rentgen.db.models.base import Base


class CustomUserProperties(Base):
    __tablename__ = "custom_user_properties"
    __table_args__ = (
        ExcludeConstraint(
            ("user_id", "="),
            ("property_id", "="),
            ("during", "&&"),
            name="uq__custom_user_properties__user_id_property_id_during",
            using="gist",
        ),
    )
    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)

    user_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("user.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
        doc="Id of corresponding user",
    )

    property_id: Mapped[int] = mapped_column(
        BigInteger,
        ForeignKey("custom_properties.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
        doc="Id of user's custom property",
    )

    value: Mapped[JSONB] = mapped_column(
        JSONB,
        nullable=False,
    )

    during: Mapped[TSTZRANGE] = mapped_column(
        TSTZRANGE,
        server_default=text("tstzrange(now(), NULL, '[)')"),
        nullable=False,
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""create_analytics_views

Revision ID: f017d4c58658
Revises: 56f5a3f9442a
Create Date: 2025-02-03 16:44:27.746785

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f017d4c58658"
down_revision = "56f5a3f9442a"
branch_labels = None
depends_on = None


def upgrade() -> None:
    for base_table in ("output", "input"):
        for depth, suffix in view_sufix_map.items():
            view_name = base_table + suffix
            op.execute(sa.text(f"DROP MATERIALIZED VIEW IF EXISTS {view_name}"))
            op.execute(sa.text(get_statement(base_table, depth, suffix)))


def downgrade() -> None:
    for base_table in ("output", "input"):
        for suffix in view_sufix_map.values():
            view_name = base_table + suffix
            op.execute(sa.text(f"DROP MATERIALIZED VIEW IF EXISTS {view_name}"))


view_sufix_map = {
    "day": "_daily_stats",
    "week": "_weekly_stats",
    "month": "_monthly_stats",
}


def get_statement(base_table: str, depth: str, suffix: str) -> str:
    view_name = base_table + suffix
    return f"""
        CREATE MATERIALIZED VIEW {view_name}
        AS (
            WITH aggregates AS (
            SELECT
                {base_table}.dataset_id as dataset_id
                , u.id as user_id
                , u.name as user_name
                , max({base_table}.created_at) as last_interaction_dt
                , count(*) as num_of_interactions
                , sum(num_bytes) as sum_bytes
                , sum(num_rows) as sum_rows
                , sum(num_files) as sum_files
                FROM {base_table}
                JOIN run r ON {base_table}.run_id = r.id
                JOIN "user" u ON r.started_by_user_id = u.id
                WHERE {base_table}.created_at >= now() - interval '1 {depth}'
                GROUP BY {base_table}.dataset_id, u.id
            )
            SELECT
                d.name as dataset_name
                , l.name as dataset_location
                , l.type as dataset_location_type
                , agr.user_id
                , agr.user_name
                , agr.last_interaction_dt
                , agr.num_of_interactions
                , agr.sum_bytes
                , agr.sum_rows
                , agr.sum_files
            FROM aggregates agr
            JOIN dataset d ON agr.dataset_id = d.id
            LEFT JOIN location l ON d.location_id = l.id
        )
        WITH NO DATA
    """



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import BigInteger, Computed, Index, String
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import Mapped, mapped_column, relationship

from data_rentgen.db.models.base import Base

if TYPE_CHECKING:
    from data_rentgen.db.models.address import Address


class Location(Base):
    """Some network location where data is bound to"""

    __tablename__ = "location"
    __table_args__ = (
        Index(None, "type", "name", unique=True),
        Index("ix__location__search_vector", "search_vector", postgresql_using="gin"),
    )

    id: Mapped[int] = mapped_column(BigInteger, primary_key=True)
    type: Mapped[str] = mapped_column(
        String(32),
        nullable=False,
        doc="Location type, e.g. kafka, postgres, hdfs",
    )
    name: Mapped[str] = mapped_column(
        String,
        nullable=False,
        doc="Location name, e.g. cluster name",
    )
    external_id: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="External ID for integration with other systems",
    )

    addresses: Mapped[list[Address]] = relationship(
        "Address",
        lazy="noload",
        back_populates="location",
    )

    search_vector: Mapped[str] = mapped_column(
        TSVECTOR,
        Computed(
            # Postgres threats host like `my.host.name` as a whole word,
            # which does not allow user to search by parts like `host.name`.
            # Keep both original name and one without punctuation to allow both full match and partial match.
            #
            # Also 'english' dictionary performs stemming,
            # so name like 'my.host.name' is converted to tsvector `'host':2 'name':3`,
            # which does not match a tsquery like 'my:* & host:* & name:*'.
            # Instead prefer 'simple' dictionary as it does not use stemming.
            "to_tsvector('simple', name || ' ' || (translate(name, '/.', ' ')) || ' ' || type)",
            persisted=True,
        ),
        nullable=False,
        deferred=True,
        doc="Full-text search vector",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

import textwrap

from pydantic import BaseModel, Field

from data_rentgen.server.settings.application_version import ApplicationVersionSettings
from data_rentgen.server.settings.cors import CORSSettings
from data_rentgen.server.settings.monitoring import MonitoringSettings
from data_rentgen.server.settings.openapi import OpenAPISettings
from data_rentgen.server.settings.request_id import RequestIDSettings
from data_rentgen.server.settings.session import SessionSettings
from data_rentgen.server.settings.static_files import StaticFilesSettings


class ServerSettings(BaseModel):
    """Data.Rentgen REST API server-specific settings.

    Examples
    --------

    .. code-block:: bash

        DATA_RENTGEN__SERVER__DEBUG=True
        DATA_RENTGEN__SERVER__MONITORING__ENABLED=True
        DATA_RENTGEN__SERVER__CORS__ENABLED=True
        DATA_RENTGEN__SERVER__REQUEST_ID__ENABLED=True
        DATA_RENTGEN__SERVER__OPENAPI__ENABLED=True
        DATA_RENTGEN__SERVER__OPENAPI__SWAGGER__ENABLED=True
        DATA_RENTGEN__SERVER__OPENAPI__REDOC__ENABLED=True
    """

    debug: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            :ref:`Enable debug output in responses <configuration-server-debug>`.
            Do not use this on production!
            """,
        ),
    )
    cors: CORSSettings = Field(
        default_factory=CORSSettings,
        description=":ref:`CORS settings <configuration-server-cors>`",
    )
    monitoring: MonitoringSettings = Field(
        default_factory=MonitoringSettings,
        description=":ref:`Monitoring settings <configuration-server-monitoring>`",
    )
    request_id: RequestIDSettings = Field(
        default_factory=RequestIDSettings,
        description=":ref:`RequestID settings <configuration-server-debug>`",
    )
    application_version: ApplicationVersionSettings = Field(
        default_factory=ApplicationVersionSettings,
        description=":ref:`Application version settings <configuration-server-debug>`",
    )
    static_files: StaticFilesSettings = Field(
        default_factory=StaticFilesSettings,
        description=":ref:`Static files settings <configuration-server-static-files>`",
    )
    openapi: OpenAPISettings = Field(
        default_factory=OpenAPISettings,
        description=":ref:`OpenAPI.json settings <configuration-server-openapi>`",
    )
    session: SessionSettings = Field(
        default_factory=SessionSettings,  # type: ignore[arg-type]
        description=":ref:`Session settings <configuration-server-session>`",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Collection
from dataclasses import dataclass
from datetime import datetime
from typing import Annotated
from uuid import UUID

from fastapi import Depends
from sqlalchemy import Row

from data_rentgen.db.models.run import Run
from data_rentgen.dto.pagination import PaginationDTO
from data_rentgen.services.uow import UnitOfWork


@dataclass
class RunServiceIOStatistics:
    total_datasets: int = 0
    total_bytes: int = 0
    total_rows: int = 0
    total_files: int = 0

    @classmethod
    def from_row(cls, row: Row | None):
        if not row:
            return cls()

        return cls(
            total_datasets=row.total_datasets,
            total_bytes=row.total_bytes or 0,
            total_rows=row.total_rows or 0,
            total_files=row.total_files or 0,
        )


@dataclass
class RunServiceOperationStatistics:
    total_operations: int = 0

    @classmethod
    def from_row(cls, row: Row | None):
        if not row:
            return cls()

        return cls(
            total_operations=row.total_operations,
        )


@dataclass
class RunServiceStatistics:
    inputs: RunServiceIOStatistics
    outputs: RunServiceIOStatistics
    operations: RunServiceOperationStatistics


@dataclass
class RunServicePageItem:
    id: UUID
    data: Run
    statistics: RunServiceStatistics


class RunServicePaginatedResult(PaginationDTO[RunServicePageItem]):
    pass


class RunService:
    def __init__(self, uow: Annotated[UnitOfWork, Depends()]):
        self._uow = uow

    async def paginate(
        self,
        page: int,
        page_size: int,
        since: datetime | None,
        until: datetime | None,
        run_ids: Collection[UUID],
        job_id: int | None,
        parent_run_id: UUID | None,
        search_query: str | None,
    ) -> RunServicePaginatedResult:
        pagination = await self._uow.run.paginate(
            page=page,
            page_size=page_size,
            since=since,
            until=until,
            run_ids=run_ids,
            job_id=job_id,
            parent_run_id=parent_run_id,
            search_query=search_query,
        )
        run_ids = [item.id for item in pagination.items]
        input_stats = await self._uow.input.get_stats_by_run_ids(run_ids)
        output_stats = await self._uow.output.get_stats_by_run_ids(run_ids)
        operation_stats = await self._uow.operation.get_stats_by_run_ids(run_ids)

        return RunServicePaginatedResult(
            page=pagination.page,
            page_size=pagination.page_size,
            total_count=pagination.total_count,
            items=[
                RunServicePageItem(
                    id=run.id,
                    data=run,
                    statistics=RunServiceStatistics(
                        inputs=RunServiceIOStatistics.from_row(input_stats.get(run.id)),
                        outputs=RunServiceIOStatistics.from_row(output_stats.get(run.id)),
                        operations=RunServiceOperationStatistics.from_row(operation_stats.get(run.id)),
                    ),
                )
                for run in pagination.items
            ],
        )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from data_rentgen.dto.column_lineage import ColumnLineageDTO
from data_rentgen.dto.dataset import DatasetDTO
from data_rentgen.dto.dataset_column_relation import (
    DatasetColumnRelationDTO,
    DatasetColumnRelationTypeDTO,
)
from data_rentgen.dto.dataset_symlink import DatasetSymlinkDTO, DatasetSymlinkTypeDTO
from data_rentgen.dto.input import InputDTO
from data_rentgen.dto.job import JobDTO, JobTypeDTO
from data_rentgen.dto.location import LocationDTO
from data_rentgen.dto.operation import (
    OperationDTO,
    OperationStatusDTO,
    OperationTypeDTO,
)
from data_rentgen.dto.output import OutputDTO, OutputTypeDTO
from data_rentgen.dto.pagination import PaginationDTO
from data_rentgen.dto.run import RunDTO, RunStartReasonDTO, RunStatusDTO
from data_rentgen.dto.schema import SchemaDTO
from data_rentgen.dto.user import UserDTO

__all__ = [
    "ColumnLineageDTO",
    "DatasetColumnRelationDTO",
    "DatasetColumnRelationTypeDTO",
    "DatasetDTO",
    "DatasetSymlinkDTO",
    "DatasetSymlinkTypeDTO",
    "InputDTO",
    "JobDTO",
    "JobTypeDTO",
    "LocationDTO",
    "OperationDTO",
    "OperationStatusDTO",
    "OperationTypeDTO",
    "OutputDTO",
    "OutputTypeDTO",
    "PaginationDTO",
    "RunDTO",
    "RunStartReasonDTO",
    "RunStatusDTO",
    "SchemaDTO",
    "UserDTO",
]



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from datetime import datetime
from enum import Enum, IntEnum
from uuid import UUID

from sqlalchemy import UUID as SQL_UUID
from sqlalchemy import (
    BigInteger,
    Computed,
    DateTime,
    Index,
    PrimaryKeyConstraint,
    SmallInteger,
    String,
)
from sqlalchemy.dialects.postgresql import TSVECTOR
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy_utils import ChoiceType

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.job import Job
from data_rentgen.db.models.user import User


class RunStatus(IntEnum):
    UNKNOWN = -1
    """No data about status"""

    STARTED = 0
    """Received START event"""

    SUCCEEDED = 1
    """Finished successfully"""

    FAILED = 2
    """Internal failure"""

    KILLED = 3
    """Killed externally, e.g. by user request or in case of OOM"""


class RunStartReason(str, Enum):
    MANUAL = "MANUAL"
    AUTOMATIC = "AUTOMATIC"

    def __str__(self) -> str:
        return str(self.value)


# no foreign keys to avoid scanning all the partitions
class Run(Base):
    __tablename__ = "run"
    __table_args__ = (
        PrimaryKeyConstraint("created_at", "id"),
        Index("ix__run__search_vector", "search_vector", postgresql_using="gin"),
        {"postgresql_partition_by": "RANGE (created_at)"},
    )

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="Timestamp component of UUID, used for table partitioning",
    )
    id: Mapped[UUID] = mapped_column(SQL_UUID)

    job_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Job the run is associated with",
    )
    job: Mapped[Job] = relationship(
        Job,
        primaryjoin="Run.job_id == Job.id",
        lazy="noload",
        foreign_keys=[job_id],
    )

    parent_run_id: Mapped[UUID | None] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=True,
        doc="Parent of current run, e.g. Airflow task run which started Spark application",
    )
    parent_run: Mapped[Run | None] = relationship(
        "Run",
        primaryjoin="Run.parent_run_id == Run.id",
        lazy="noload",
        foreign_keys=[parent_run_id],
    )

    status: Mapped[RunStatus] = mapped_column(
        ChoiceType(RunStatus, impl=SmallInteger()),
        nullable=False,
        default=RunStatus.UNKNOWN,
        doc="Run status info",
    )

    external_id: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="External ID of the run, e.g. Spark applicationId",
    )
    attempt: Mapped[str | None] = mapped_column(
        String(64),
        nullable=True,
        doc="Attempt number of the run",
    )
    persistent_log_url: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="Persistent log url of the run, like Spark history server url, optional",
    )
    running_log_url: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="Log url of the run in progress, like Spark session UI url, optional",
    )

    started_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        doc="Start time of the run",
    )
    started_by_user_id: Mapped[int | None] = mapped_column(
        BigInteger,
        nullable=True,
        doc="User who started the run",
    )
    started_by_user: Mapped[User | None] = relationship(
        "User",
        primaryjoin="Run.started_by_user_id == User.id",
        lazy="noload",
        foreign_keys=[started_by_user_id],
    )
    start_reason: Mapped[RunStartReason | None] = mapped_column(
        ChoiceType(RunStartReason, impl=String(32)),
        nullable=True,
        doc="Start reason of the run, e.g. MANUAL or AUTOMATIC",
    )

    ended_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=True),
        nullable=True,
        doc="End time of the run",
    )
    end_reason: Mapped[str | None] = mapped_column(
        String,
        nullable=True,
        doc="End reason of the run, e.g. exception string",
    )

    search_vector: Mapped[str] = mapped_column(
        TSVECTOR,
        Computed(
            # Postgres treats values like `my.app.id` as a whole word (domain name),
            # which does not allow user to search by parts like `some` or `value`.
            # Same for slashes which are treated like file paths.
            # Keep both original name and one without punctuation to allow both full match and partial match.
            #
            # Also 'english' dictionary performs stemming,
            # so name like 'my.app.id' is converted to tsvector `'app':2 'id':3`,
            # which does not match a tsquery like 'my:* & app:* & id:*'.
            # Instead prefer 'simple' dictionary as it does not use stemming.
            "to_tsvector('simple'::regconfig, external_id || ' ' || (translate(external_id, '/.', '  ')))",
            persisted=True,
        ),
        nullable=True,
        deferred=True,
        doc="Full-text search vector",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Remove partitions from dataset names

Revision ID: 976168ee4f16
Revises: 15c0a22b8566
Create Date: 2025-03-17 18:28:24.545234

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "976168ee4f16"
down_revision = "15c0a22b8566"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # Create a Map old_id -> new_id
    op.execute(sa.text("CREATE TEMP TABLE dataset_migration (id SERIAL PRIMARY KEY, old_id INT, new_id INT);"))
    # Add new rows into dataset with 'short' name if such names doesn't exists
    op.execute(
        sa.text(
            """
            INSERT INTO dataset (name, location_id, format)
            SELECT DISTINCT REGEXP_REPLACE(name, '/[^/]*=[^/]*.*', ''), location_id, format
            FROM dataset WHERE regexp_match(name, '/[^/]*=[^/]*.*') is not null
            ON CONFLICT DO NOTHING;
            """,
        ),
    )
    # Fill Map table
    op.execute(
        sa.text(
            """
            WITH new_datasets as (
                SELECT id, name
                FROM dataset WHERE name in (
                    SELECT distinct regexp_replace(name, '/[^/]*=[^/]*.*', '')
                    FROM dataset WHERE regexp_match(name, '/[^/]*=[^/]*.*') is not null)
            )
            INSERT INTO dataset_migration (old_id, new_id)
            SELECT d1.id, d2.id
            FROM dataset AS d1
            JOIN new_datasets AS d2 ON REGEXP_REPLACE(d1.name, '/[^/]*=[^/]*.*', '') = d2.name;
            """,
        ),
    )
    # Update dataset_ids in related tables
    op.execute(
        sa.text(
            """
            UPDATE output o
            SET dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE o.dataset_id = dm.old_id;
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            UPDATE input o
            SET dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE o.dataset_id = dm.old_id;
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            UPDATE column_lineage cl
            SET source_dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE cl.source_dataset_id = dm.old_id;
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            UPDATE column_lineage cl
            SET target_dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE cl.target_dataset_id = dm.old_id;
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            UPDATE dataset_symlink ds
            SET from_dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE ds.from_dataset_id = dm.old_id;
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            UPDATE dataset_symlink ds
            SET to_dataset_id = dm.new_id
            FROM dataset_migration dm
            WHERE ds.to_dataset_id = dm.old_id
            """,
        ),
    )
    # Delete row with old datasets
    op.execute(
        sa.text(
            """
            DELETE FROM dataset
            WHERE id IN (SELECT old_id FROM dataset_migration);
            """,
        ),
    )
    # Remove duplicates from input and output
    op.execute(
        sa.text(
            """
            WITH duplicates as (
                SELECT id, row_number() over(partition by dataset_id, operation_id order by created_at desc) as row_num
                FROM input
            )
            DELETE from input where id in (select id from duplicates where row_num > 1);
            """,
        ),
    )
    op.execute(
        sa.text(
            """
            WITH duplicates as (
                SELECT id, row_number() over(partition by dataset_id, operation_id order by created_at desc) as row_num
                FROM output
            )
            DELETE from output where id in (select id from duplicates where row_num > 1);
            """,
        ),
    )


def downgrade() -> None:
    # This migration is irreversible
    pass



# SPDX-FileCopyrightText: 2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
"""Create custom user properties

Revision ID: 56f5a3f9442a
Revises: 61ea5edad711
Create Date: 2025-01-09 13:25:39.718162

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB, TSTZRANGE, ExcludeConstraint

# revision identifiers, used by Alembic.
revision = "56f5a3f9442a"
down_revision = "61ea5edad711"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(sa.text("CREATE EXTENSION IF NOT EXISTS btree_gist;"))
    op.create_table(
        "custom_user_properties",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.Column("user_id", sa.BigInteger, nullable=False),
        sa.Column("property_id", sa.BigInteger, nullable=False),
        sa.Column("value", JSONB, nullable=False),
        sa.Column(
            "during",
            TSTZRANGE,
            server_default=sa.text("tstzrange(now(), NULL, '[)')"),
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk__custom_user_properties")),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk__custom_user_properties__user_id"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["property_id"],
            ["custom_properties.id"],
            name=op.f("fk__custom_user_properties__custom_properties_id"),
            ondelete="CASCADE",
        ),
        ExcludeConstraint(
            ("user_id", "="),
            ("property_id", "="),
            ("during", "&&"),
            name="uq__custom_user_properties__user_id_property_id_during",
            using="gist",
        ),
    )
    op.create_index(
        op.f("ix__custom_user_properties__user_id"),
        "custom_user_properties",
        ["user_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix__custom_user_properties__property_id"),
        "custom_user_properties",
        ["property_id"],
        unique=False,
    )


def downgrade() -> None:
    op.drop_index(
        op.f("ix__custom_user_properties__user_id"),
        "custom_user_properties",
    )
    op.drop_index(
        op.f("ix__custom_user_properties__property_id"),
        "custom_user_properties",
    )
    op.drop_table("custom_user_properties")



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from datetime import datetime
from enum import Enum
from uuid import UUID

from sqlalchemy import UUID as SQL_UUID
from sqlalchemy import BigInteger, DateTime, PrimaryKeyConstraint, String
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy_utils import ChoiceType

from data_rentgen.db.models.base import Base
from data_rentgen.db.models.dataset import Dataset
from data_rentgen.db.models.job import Job
from data_rentgen.db.models.operation import Operation
from data_rentgen.db.models.run import Run
from data_rentgen.db.models.schema import Schema


class OutputType(str, Enum):
    CREATE = "CREATE"
    ALTER = "ALTER"
    RENAME = "RENAME"

    APPEND = "APPEND"
    OVERWRITE = "OVERWRITE"

    DROP = "DROP"
    TRUNCATE = "TRUNCATE"

    def __str__(self) -> str:
        return self.value


# no foreign keys to avoid scanning all the partitions
class Output(Base):
    __tablename__ = "output"
    __table_args__ = (
        PrimaryKeyConstraint("created_at", "id"),
        {"postgresql_partition_by": "RANGE (created_at)"},
    )

    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="Timestamp component of UUID, used for table partitioning",
    )
    id: Mapped[UUID] = mapped_column(SQL_UUID)

    operation_id: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=False,
        doc="Operation caused this output",
    )
    operation: Mapped[Operation] = relationship(
        Operation,
        primaryjoin="Output.operation_id == Operation.id",
        lazy="noload",
        foreign_keys=[operation_id],
    )

    run_id: Mapped[UUID] = mapped_column(
        SQL_UUID,
        index=True,
        nullable=False,
        doc="Run caused this input operation",
    )
    run: Mapped[Run] = relationship(
        Run,
        primaryjoin="Output.run_id == Run.id",
        lazy="noload",
        foreign_keys=[run_id],
    )

    job_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Parent job of run",
    )
    job: Mapped[Job] = relationship(
        Job,
        primaryjoin="Output.job_id == Job.id",
        lazy="noload",
        foreign_keys=[job_id],
    )

    dataset_id: Mapped[int] = mapped_column(
        BigInteger,
        index=True,
        nullable=False,
        doc="Dataset the output is performed against",
    )
    dataset: Mapped[Dataset] = relationship(
        Dataset,
        primaryjoin="Output.dataset_id == Dataset.id",
        lazy="noload",
        foreign_keys=[dataset_id],
    )

    type: Mapped[OutputType] = mapped_column(
        ChoiceType(OutputType, impl=String(32)),
        nullable=False,
        default=OutputType.APPEND,
        doc="Type of the output, e.g. READ, CREATE, APPEND",
    )

    schema_id: Mapped[int | None] = mapped_column(
        BigInteger,
        index=True,
        nullable=True,
        doc="Schema the output is performed with, if any",
    )
    schema: Mapped[Schema | None] = relationship(
        Schema,
        primaryjoin="Output.schema_id == Schema.id",
        lazy="noload",
        foreign_keys=[schema_id],
    )

    num_bytes: Mapped[int | None] = mapped_column(
        BigInteger,
        nullable=True,
        doc="Amount of data moved during Output, in bytes",
    )

    num_rows: Mapped[int | None] = mapped_column(
        BigInteger,
        nullable=True,
        doc="Amount of data moved during Output, in rows",
    )

    num_files: Mapped[int | None] = mapped_column(
        BigInteger,
        nullable=True,
        doc="Amount of data moved during Output, in rows",
    )



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from fastapi import APIRouter

from data_rentgen.server.schemas import PingResponse

router = APIRouter(tags=["Monitoring"], prefix="/monitoring")


@router.get("/ping", summary="Check if server is alive")
async def ping() -> PingResponse:
    return PingResponse()



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends, Query

from data_rentgen.db.models import User
from data_rentgen.server.errors import get_error_responses
from data_rentgen.server.errors.schemas import InvalidRequestSchema
from data_rentgen.server.schemas.v1 import (
    JobDetailedResponseV1,
    JobLineageQueryV1,
    JobPaginateQueryV1,
    LineageResponseV1,
    PageResponseV1,
)
from data_rentgen.server.services import JobService, LineageService, get_user
from data_rentgen.server.utils.lineage_response import build_lineage_response

router = APIRouter(prefix="/jobs", tags=["Jobs"], responses=get_error_responses(include={InvalidRequestSchema}))


@router.get("", summary="Paginated list of Jobs")
async def paginate_jobs(
    query_args: Annotated[JobPaginateQueryV1, Depends()],
    job_service: Annotated[JobService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> PageResponseV1[JobDetailedResponseV1]:
    pagination = await job_service.paginate(
        page=query_args.page,
        page_size=query_args.page_size,
        job_ids=query_args.job_id,
        search_query=query_args.search_query,
    )
    return PageResponseV1[JobDetailedResponseV1].from_pagination(pagination)


@router.get("/lineage", summary="Get Job lineage graph")
async def get_jobs_lineage(
    query_args: Annotated[JobLineageQueryV1, Query()],
    lineage_service: Annotated[LineageService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> LineageResponseV1:
    lineage = await lineage_service.get_lineage_by_jobs(
        start_node_ids=[query_args.start_node_id],  # type: ignore[list-item]
        direction=query_args.direction,
        granularity=query_args.granularity,
        since=query_args.since,
        until=query_args.until,
        depth=query_args.depth,
        include_column_lineage=query_args.include_column_lineage,
    )

    return await build_lineage_response(lineage)



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import datetime
from enum import IntEnum
from uuid import UUID

from fastapi import Query
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_serializer,
    field_validator,
    model_validator,
)

from data_rentgen.server.schemas.v1.pagination import PaginateQueryV1
from data_rentgen.server.schemas.v1.user import UserResponseV1
from data_rentgen.utils import UUIDv6Plus


class RunStatusV1(IntEnum):
    UNKNOWN = -1
    """No data about status"""

    STARTED = 0
    """Received START event"""

    SUCCEEDED = 1
    """Finished successfully"""

    FAILED = 2
    """Internal failure"""

    KILLED = 3
    """Killed externally, e.g. by user request or in case of OOM"""

    def __str__(self) -> str:
        return self.name


class RunResponseV1(BaseModel):
    """Run response"""

    id: UUID = Field(description="Run id")
    created_at: datetime = Field(description="Run creation time")
    job_id: str = Field(description="Job the run is associated with", coerce_numbers_to_str=True)
    parent_run_id: UUID | None = Field(description="Parent of current run", default=None)
    status: RunStatusV1 = Field(description="Run status")
    external_id: str | None = Field(description="External id, e.g. Spark applicationid", default=None)
    attempt: str | None = Field(description="Attempt number of the run", default=None)
    persistent_log_url: str | None = Field(
        description="Persistent log url of the run, like Spark history server url, optional",
        default=None,
    )
    running_log_url: str | None = Field(
        description="Log url of the run in progress, like Spark session UI url, optional",
        default=None,
    )
    started_at: datetime | None = Field(description="Start time of the Run", default=None)
    started_by_user: UserResponseV1 | None = Field(description="User who started the Run", default=None)
    start_reason: str | None = Field(description="Start reason of the Run", default=None)
    ended_at: datetime | None = Field(description="End time of the Run", default=None)
    end_reason: str | None = Field(description="End reason of the Run", default=None)

    model_config = ConfigDict(from_attributes=True)

    @field_serializer("status", when_used="json-unless-none")
    def _serialize_status(self, value: RunStatusV1) -> str:
        return str(value)


class RunIOStatisticsReponseV1(BaseModel):
    """Run IO statistics response."""

    total_datasets: int = Field(default=0, description="Total number of datasets")
    total_bytes: int = Field(default=0, description="Total number of bytes")
    total_rows: int = Field(default=0, description="Total number of rows")
    total_files: int = Field(default=0, description="Total number of files")

    model_config = ConfigDict(from_attributes=True)


class RunOperationStatisticsReponseV1(BaseModel):
    """Run operation statistics response."""

    total_operations: int = Field(default=0, description="Total number of operations")

    model_config = ConfigDict(from_attributes=True)


class RunStatisticsReponseV1(BaseModel):
    """Run statistics response."""

    outputs: RunIOStatisticsReponseV1 = Field(description="Output statistics")
    inputs: RunIOStatisticsReponseV1 = Field(description="Input statistics")
    operations: RunOperationStatisticsReponseV1 = Field(description="Operation statistics")

    model_config = ConfigDict(from_attributes=True)


class RunDetailedResponseV1(BaseModel):
    """Run response."""

    id: UUID = Field(description="Run id")
    data: RunResponseV1 = Field(description="Run data")
    statistics: RunStatisticsReponseV1 = Field(description="Run statistics")

    model_config = ConfigDict(from_attributes=True)


class RunsQueryV1(PaginateQueryV1):
    """Query params for Runs paginate request."""

    since: datetime | None = Field(
        Query(
            default=None,
            description="Minimum value of Run 'created_at' field, in ISO 8601 format",
            examples=["2008-09-15T15:53:00+05:00"],
        ),
    )
    until: datetime | None = Field(
        Query(
            default=None,
            description="Maximum value of Run 'created_at' field, in ISO 8601 format",
            examples=["2008-09-15T15:53:00+05:00"],
        ),
    )
    run_id: list[UUIDv6Plus] = Field(
        Query(
            default_factory=list,
            description="Run ids, for exact match",
        ),
    )
    job_id: int | None = Field(
        Query(
            default=None,
            description="Job id, can be used only with 'since'",
        ),
    )

    parent_run_id: UUIDv6Plus | None = Field(
        Query(
            default=None,
            description="Parent run id, can be used only with 'since' and 'until'",
            examples=["01913217-b761-7b1a-bb52-489da9c8b9c8"],
        ),
    )

    search_query: str | None = Field(
        Query(
            default=None,
            min_length=3,
            description="Search query",
        ),
    )

    model_config = ConfigDict(extra="forbid")

    @field_validator("until", mode="after")
    @classmethod
    def _check_until(cls, value: datetime | None, info: ValidationInfo) -> datetime | None:
        since = info.data.get("since")
        if since and value and since >= value:
            msg = "'since' should be less than 'until'"
            raise ValueError(msg)
        return value

    @model_validator(mode="after")
    def _check_fields(self):
        if not any([self.since, self.run_id, self.job_id, self.parent_run_id, self.search_query]):
            msg = "input should contain either 'since', 'run_id', 'job_id', 'parent_run_id' or 'search_query' field"
            raise ValueError(msg)
        if self.job_id and not self.since:
            msg = "'job_id' can be passed only with 'since'"
            raise ValueError(msg)
        if self.parent_run_id and not self.since:
            msg = "'parent_run_id' can be passed only with 'since'"
            raise ValueError(msg)
        if self.search_query and not self.since:
            msg = "'search_query' can be passed only with 'since'"
            raise ValueError(msg)
        return self



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Collection
from datetime import datetime, timezone
from uuid import UUID

from sqlalchemy import Row, UnaryExpression, any_, func, select
from sqlalchemy.dialects.postgresql import insert

from data_rentgen.db.models import Operation, OperationStatus, OperationType
from data_rentgen.db.repositories.base import Repository
from data_rentgen.db.utils.uuid import extract_timestamp_from_uuid
from data_rentgen.dto import OperationDTO, PaginationDTO


class OperationRepository(Repository[Operation]):
    async def create_or_update_bulk(self, operations: list[OperationDTO]) -> None:
        if not operations:
            return

        insert_statement = insert(Operation)
        statement = insert_statement.on_conflict_do_update(
            index_elements=[Operation.created_at, Operation.id],
            set_={
                "name": func.coalesce(insert_statement.excluded.name, Operation.name),
                "type": func.coalesce(insert_statement.excluded.type, Operation.type),
                "status": func.greatest(insert_statement.excluded.status, Operation.status),
                "started_at": func.coalesce(insert_statement.excluded.started_at, Operation.started_at),
                "ended_at": func.coalesce(insert_statement.excluded.ended_at, Operation.ended_at),
                "description": func.coalesce(insert_statement.excluded.description, Operation.description),
                "group": func.coalesce(insert_statement.excluded.group, Operation.group),
                "position": func.coalesce(insert_statement.excluded.position, Operation.position),
            },
        )

        await self._session.execute(
            statement,
            [
                {
                    "id": operation.id,
                    "created_at": extract_timestamp_from_uuid(operation.id),
                    "run_id": operation.run.id,
                    "name": operation.name,
                    "type": OperationType(operation.type) if operation.type else None,
                    "status": OperationStatus(operation.status),
                    "started_at": operation.started_at,
                    "ended_at": operation.ended_at,
                    "description": operation.description,
                    "group": operation.group,
                    "position": operation.position,
                }
                for operation in operations
            ],
        )

    async def paginate(
        self,
        page: int,
        page_size: int,
        operation_ids: Collection[UUID],
        since: datetime | None,
        until: datetime | None,
        run_id: UUID | None,
    ) -> PaginationDTO[Operation]:
        # do not use `tuple_(Operation.created_at, Operation.id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        where = []
        if operation_ids:
            min_operation_created_at = extract_timestamp_from_uuid(min(operation_ids))
            max_operation_created_at = extract_timestamp_from_uuid(max(operation_ids))
            min_created_at = max(since, min_operation_created_at) if since else min_operation_created_at
            max_created_at = min(until, max_operation_created_at) if until else max_operation_created_at
            where = [
                Operation.created_at >= min_created_at,
                Operation.created_at <= max_created_at,
            ]
        else:
            if since:
                where.append(Operation.created_at >= since)
            if until:
                where.append(Operation.created_at <= until)

        if run_id:
            where.append(Operation.run_id == run_id)
        if operation_ids:
            where.append(Operation.id == any_(list(operation_ids)))  # type: ignore[arg-type]

        query = select(Operation).where(*where)
        order_by: list[UnaryExpression] = [Operation.created_at.desc(), Operation.id.desc()]
        return await self._paginate_by_query(
            query=query,
            order_by=order_by,
            page=page,
            page_size=page_size,
        )

    async def list_by_run_ids(
        self,
        run_ids: Collection[UUID],
        since: datetime,
        until: datetime | None,
    ) -> list[Operation]:
        if not run_ids:
            return []

        # All operations are created after run
        min_run_created_at = extract_timestamp_from_uuid(min(run_ids))
        min_operation_created_at = max(min_run_created_at, since.astimezone(timezone.utc))
        query = select(Operation).where(
            Operation.created_at >= min_operation_created_at,
            Operation.run_id == any_(list(run_ids)),  # type: ignore[arg-type]
        )
        if until:
            query = query.where(Operation.created_at <= until)
        result = await self._session.scalars(query)
        return list(result.all())

    async def list_by_ids(self, operation_ids: Collection[UUID]) -> list[Operation]:
        if not operation_ids:
            return []
        # do not use `tuple_(Operation.created_at, Operation.id).in_(...),
        # as this is too complex filter for Postgres to make an optimal query plan
        min_created_at = extract_timestamp_from_uuid(min(operation_ids))
        max_created_at = extract_timestamp_from_uuid(max(operation_ids))
        query = select(Operation).where(
            Operation.created_at >= min_created_at,
            Operation.created_at <= max_created_at,
            Operation.id == any_(list(operation_ids)),  # type: ignore[arg-type]
        )
        result = await self._session.scalars(query)
        return list(result.all())

    async def get_stats_by_run_ids(self, run_ids: Collection[UUID]) -> dict[UUID, Row]:
        if not run_ids:
            return {}

        # unlike list_by_run_ids, we need to get all statistics for specific runs, regardless of time range
        min_created_at = extract_timestamp_from_uuid(min(run_ids))
        query = (
            select(
                Operation.run_id.label("run_id"),
                func.count(Operation.id.distinct()).label("total_operations"),
            )
            .where(
                Operation.created_at >= min_created_at,
                Operation.run_id == any_(list(run_ids)),  # type: ignore[arg-type]
            )
            .group_by(Operation.run_id)
        )

        query_result = await self._session.execute(query)
        return {row.run_id: row for row in query_result.all()}



# SPDX-FileCopyrightText: 2024-2025 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from typing import Annotated

from fastapi import APIRouter, Depends, Query

from data_rentgen.db.models import User
from data_rentgen.server.errors import get_error_responses
from data_rentgen.server.errors.schemas import InvalidRequestSchema
from data_rentgen.server.schemas.v1 import (
    LineageResponseV1,
    OperationDetailedResponseV1,
    OperationLineageQueryV1,
    OperationQueryV1,
    PageResponseV1,
)
from data_rentgen.server.services import get_user
from data_rentgen.server.services.lineage import LineageService
from data_rentgen.server.services.operation import OperationService
from data_rentgen.server.utils.lineage_response import build_lineage_response
from data_rentgen.services import UnitOfWork

router = APIRouter(
    prefix="/operations",
    tags=["Operations"],
    responses=get_error_responses(include={InvalidRequestSchema}),
)


# TODO: Remove type ignore in DOP-24446
@router.get("", summary="Paginated list of Operations")
async def operations(
    query_args: Annotated[OperationQueryV1, Depends()],
    unit_of_work: Annotated[UnitOfWork, Depends()],
    operation_service: Annotated[OperationService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> PageResponseV1[OperationDetailedResponseV1]:
    pagination = await operation_service.paginate(
        page=query_args.page,
        page_size=query_args.page_size,
        since=query_args.since,
        until=query_args.until,
        operation_ids=query_args.operation_id,
        run_id=query_args.run_id,
    )
    return PageResponseV1[OperationDetailedResponseV1].from_pagination(pagination)


@router.get("/lineage", summary="Get Operation lineage graph")
async def get_operations_lineage(
    query_args: Annotated[OperationLineageQueryV1, Query()],
    lineage_service: Annotated[LineageService, Depends()],
    current_user: Annotated[User, Depends(get_user())],
) -> LineageResponseV1:
    lineage = await lineage_service.get_lineage_by_operations(
        start_node_ids=[query_args.start_node_id],
        direction=query_args.direction,
        since=query_args.since,
        until=query_args.until,
        depth=query_args.depth,
        include_column_lineage=query_args.include_column_lineage,
    )

    return await build_lineage_response(lineage)
